#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "avifile.h"

struct avifile_list_node
{
    int type;
    int idx1_len;
    void *avih,*strh,*strf,*strn,*strd,*indx,*dmlh,*junk,*idx1;

    int strl_count;
    struct avifile_list_node** strl;
    
    struct avifile_list_node *hdrl, *movi, *odml;
    
    // for debugs....
    int fpos;
};

struct avifile_idx1_item
{
    int ckid;
    int Flags;
    int ChunkOffset;
    int ChunkLength;
};


void avifile_dump_node(void* n,long level)
{
    // prepare offset string
    char offset[120]; 
    memset(offset,0,sizeof(offset));
    memset(offset,'\t',level);
    
    // vars
    int i;

    // cast node    
    struct avifile_list_node* node = (struct avifile_list_node*)n;
    
    if(node)
    {
        // static ptrs
        printf("%stype: %.8X [%c%c%c%c]\n",offset,node->type, ((char*)&node->type)[0], ((char*)&node->type)[1], ((char*)&node->type)[2], ((char*)&node->type)[3] );
        printf("%sidx1_len: %.8X\n",offset,node->idx1_len);
        printf("%savih: %.8X\n",offset,node->avih );
        printf("%sstrh: %.8X\n",offset,node->strh );
        printf("%sstrf: %.8X\n",offset,node->strf );
        printf("%sstrn: %.8X\n",offset,node->strn );
        printf("%sstrd: %.8X\n",offset,node->strd );
        printf("%sindx: %.8X\n",offset,node->indx );
        printf("%sdmlh: %.8X\n",offset,node->dmlh );
        printf("%sjunk: %.8X\n",offset,node->junk );
        printf("%sidx1: %.8X\n",offset,node->idx1 );

        // leaves
        printf("%shdrl: %.8X\n",offset,node->hdrl ); avifile_dump_node(node->hdrl, level +1);
        printf("%smovi: %.8X\n",offset,node->movi ); avifile_dump_node(node->movi, level +1);
        printf("%sodml: %.8X\n",offset,node->odml ); avifile_dump_node(node->odml, level +1);
        
        // streams list
        for(i=0;i<node->strl_count;i++)
            printf("%sstrl[%3d]: %.8X\n",offset,i,node->strl[i]); avifile_dump_node(node->strl[i], level +1);
    };
        
};

//#define AVIFILE_VERBOSE "avifile: "

#define FOURCC_CONST_RIFF FOURCC_DIRECT('R','I','F','F')
#define FOURCC_CONST_LIST FOURCC_DIRECT('L','I','S','T')
#define FOURCC_CONST_AVI  FOURCC_DIRECT('A','V','I',' ')
#define FOURCC_CONST_SPACE  FOURCC_DIRECT(' ',' ',' ',' ')
#define FOURCC_CONST_HDRL FOURCC_DIRECT('h','d','r','l')
#define FOURCC_CONST_AVIH FOURCC_DIRECT('a','v','i','h')
#define FOURCC_CONST_STRL FOURCC_DIRECT('s','t','r','l')
#define FOURCC_CONST_STRH FOURCC_DIRECT('s','t','r','h')
#define FOURCC_CONST_STRN FOURCC_DIRECT('s','t','r','n')
#define FOURCC_CONST_STRD FOURCC_DIRECT('s','t','r','d')
#define FOURCC_CONST_STRF FOURCC_DIRECT('s','t','r','f')
#define FOURCC_CONST_MOVI FOURCC_DIRECT('m','o','v','i')
#define FOURCC_CONST_INDX FOURCC_DIRECT('i','n','d','x')
#define FOURCC_CONST_JUNK FOURCC_DIRECT('J','U','N','K')
#define FOURCC_CONST_ODML FOURCC_DIRECT('o','d','m','l')
#define FOURCC_CONST_DMLH FOURCC_DIRECT('d','m','l','h')
#define FOURCC_CONST_IDX1 FOURCC_DIRECT('i','d','x','1')


#define FOURCC_CONST_VIDS FOURCC_DIRECT('v','i','d','s')
#define FOURCC_CONST_AUDS FOURCC_DIRECT('a','u','d','s')
#define FOURCC_CONST_IVAS FOURCC_DIRECT('i','v','a','s')


#define FOURCC_DEBUG_CHARS(VAL) ((unsigned char*)&VAL)[0], ((unsigned char*)&VAL)[1], ((unsigned char*)&VAL)[2], ((unsigned char*)&VAL)[3]

#define CLEAR(x) memset (&(x), 0, sizeof (x))

//#define DEBUG_ASIB

// thus function build subcjunk index
struct avifile_subchunk_item* avifile_subchunks_index_build(struct avifile_object* ao, int subchunk_size, long subchunk_count, void* _idx, int idx_len)
{
//    struct avifile_idx1_item* idx1;

    // manual cast
    struct avifile_idx1_item* idx1 = (struct avifile_idx1_item*)(((struct avifile_list_node*)(ao->riff))->idx1);
    int* idx = (int*)_idx;

    struct avifile_subchunk_item* subchunk_index = (struct avifile_subchunk_item*)malloc(subchunk_count * sizeof(struct avifile_subchunk_item) );
    bzero
    (
	subchunk_index,
	subchunk_count * sizeof(struct avifile_subchunk_item) 
    );

#ifdef DEBUG_ASIB
    FILE* f= fopen("_avifile_subchunks_index_build","wt");
    fprintf(f,"subchunk_size=%d, subchunk_count=%ld, idx1=%.8lX, idx1_len=%d\n", subchunk_size, subchunk_count, (long)idx, idx_len);
#endif

    for(int i=0,j=0;(i<idx_len)&&(j<subchunk_count);i++)
    {
	int c = ((struct avifile_idx1_item*)idx1)[ idx[i] ].ChunkLength / subchunk_size;
#ifdef DEBUG_ASIB
	fprintf(f,"i=%d,j=%d,c=%d, idx1[%d].ChunkLength=%d\n",i,j,c, idx[i],  ((struct avifile_idx1_item*)idx1)[ idx[i] ].ChunkLength);
#endif
	for(int k = 0; k < c; k++,j++)
	{
	    subchunk_index[j].chunk_i = i;
	    subchunk_index[j].offset = k;
#ifdef DEBUG_ASIB
	    fprintf(f,"\tsybchunk[%d]={chunk_i=%ld, offset=%ld}\n",j,subchunk_index[j].chunk_i,subchunk_index[j].offset);
#endif
	};
    };

#ifdef DEBUG_ASIB
    fclose(f);
#endif

    return subchunk_index;
};

void avifile_subchunks_index_free(struct avifile_subchunk_item* p)
{
    free(p);
};

// loads chunk
void* avifile_chunk_load(struct avifile_object* ao, int stream, int num)
{
    struct avifile_list_node* node;
    struct avifile_idx1_item* idx1;
    int i;
    void* chunk;
    int buf2[2];
    
    if((ao)&&(ao->idx1[stream])&&(ao->riff))
    {
        // assign node
        node = (struct avifile_list_node*)ao->riff;
        
        // assign 'idx1' table
        idx1 = (struct avifile_idx1_item*)node->idx1;

        // check if desired chunk is not outside
        if(num < ao->idx1_count[stream])
        {
            // determinate index of chunk
            i = ao->idx1[stream][num];

#ifdef AVIFILE_VERBOSE
            printf(AVIFILE_VERBOSE "NOTICE(avifile_chunk_load): ENTER: i=%d,stream=%d,num=%d,offset=%.8X\n",i,stream,num,idx1[i].ChunkOffset);
#endif

            // seek to position
            fseek((FILE*)ao->file,idx1[i].ChunkOffset,SEEK_SET);


            // allocate space
            chunk = malloc(idx1[i].ChunkLength + 8);

            
            // load
            fread(chunk,1,idx1[i].ChunkLength + 8,(FILE*)ao->file);

#ifdef AVIFILE_VERBOSE
            printf(AVIFILE_VERBOSE "NOTICE(avifile_chunk_load): CHUNK: len=%d, TYPE=('%c','%c','%c','%c') \n", (((unsigned int*)chunk)[1]), FOURCC_DEBUG_CHARS( (((unsigned int*)chunk)[1]) ));
#endif

            // move forward in 8 bytes
            return (((unsigned char*)chunk) + 8 );
        };
        
    };
    
    return NULL;
};


int avifile_chunk_length(struct avifile_object* ao, int stream, int num)
{
    struct avifile_list_node* node;
    struct avifile_idx1_item* idx1;
    int i;
    void* chunk;
    int buf2[2];
    
    if((ao)&&(ao->idx1[stream])&&(ao->riff))
    {
        // assign node
        node = (struct avifile_list_node*)ao->riff;
        
        // assign 'idx1' table
        idx1 = (struct avifile_idx1_item*)node->idx1;

        // check if desired chunk is not outside
        if(num < ao->idx1_count[stream])
        {
            // determinate index of chunk
            i = ao->idx1[stream][num];

            return idx1[i].ChunkLength;
        };
        
    };
    return 0;
};



void avifile_chunk_free(void* chunk)
{
    if(!(chunk))
	return;

    unsigned char* origin_chunk = (unsigned char*)chunk;
    origin_chunk -= 8;
    
    // no check yet
    free(origin_chunk);
    

//    free(chunk);
};



static void avifile_list_node_free(struct avifile_list_node** node)
{
#define FREE1(VAR) if((*node)->VAR) free((*node)->VAR);(*node)->VAR = NULL;
#define FREE2(VAR) if((*node)->VAR) avifile_list_node_free(&(*node)->VAR);

    int i;
    
    FREE1(avih);FREE1(strh);FREE1(strf);FREE1(strn);FREE1(strd);FREE1(indx);FREE1(dmlh);FREE1(junk);
    FREE2(hdrl);FREE2(movi);FREE2(odml);
    
    for(i=0;i<(*node)->strl_count;i++) FREE2(strl[i]);
    
    FREE1(strl);

    FREE1(idx1);
    
    free(*node);
    *node = NULL;
};

void avifile_close(struct avifile_object* ao)
{
    int i;
    
    // check if object is OK
    if(!(ao))
        return;
        
    // find 'node'
    struct avifile_list_node* node = (struct avifile_list_node*)ao->riff;

    if(ao->file)
	fclose((FILE*)ao->file);
	
    avifile_list_node_free(&node);
    
    for(i=0;i<AVIFILE_MAXIMUM_STREAMS;i++)
        free(ao->idx1[i]);
    
    free(ao);

};


static struct avifile_list_node* avifile_list_node_load(FILE* file,int type,int parse_level,int block_limit)
{
    // alloc node
    struct avifile_list_node* node = (struct avifile_list_node*)malloc(sizeof(struct avifile_list_node));
    struct avifile_list_node* last_block = NULL;
    
    // set flag
    int ok = 1;
    
    // buffer for reading
    unsigned int buf3[3];
    unsigned int buf2[2];

    // aux vars
    void **ptr;
    char* blockname;

    // clear mem of struct
    CLEAR(*node);

    // save position
    node->fpos = ftell(file);

#ifdef AVIFILE_VERBOSE
    printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): ENTER: LEVEL=%d, FPOS=%.8lX, TYPE=('%c','%c','%c','%c') \n", parse_level, ftell(file), FOURCC_DEBUG_CHARS(type));
#endif
        
    switch(type)
    {
	case FOURCC_CONST_AVI:
#ifdef AVIFILE_VERBOSE
	    printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): type = FOURCC_CONST_AVI \n");
#endif
	    break;




	case FOURCC_CONST_MOVI:
#ifdef AVIFILE_VERBOSE
	    printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): type = FOURCC_CONST_MOVI \n");
#endif
	    // return node
	    return node;
	    
	    break;
	    
	case FOURCC_CONST_STRL:
#ifdef AVIFILE_VERBOSE
	    printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): type = FOURCC_CONST_STRL \n");
#endif
	
	    // endless loop for reading data
	    while(1)
	    {
#ifdef AVIFILE_VERBOSE
		printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): FPOS=%.8lX \n",ftell(file));
#endif
		// load 'strh/strf/strd/strn' header
		if( sizeof(buf2) != fread(buf2,1,sizeof(buf2),file) )
		{
#ifdef AVIFILE_VERBOSE
		    printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): unable to read 'strh/strf/strd/strn' header\n");
#endif
		    free(node);
		    return NULL;
		};
	    
		// determinate what header is loaded
		switch (buf2[0])
		{
		    case FOURCC_CONST_STRH: ptr = &node->strh;blockname = "strh";break;
		    case FOURCC_CONST_STRF: ptr = &node->strf;blockname = "strf";break;
		    case FOURCC_CONST_STRD: ptr = &node->strd;blockname = "strd";break;
		    case FOURCC_CONST_STRN: ptr = &node->strn;blockname = "strn";break;
		    case FOURCC_CONST_INDX: ptr = &node->strn;blockname = "indx";break;
		    case FOURCC_CONST_JUNK: ptr = &node->junk;blockname = "JUNK";break;
		    default:
			// not recognized !!
			// roll back
			fseek(file,-sizeof(buf2),SEEK_CUR);
			// return struct
#ifdef AVIFILE_VERBOSE
			printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): no more 'strh/strf/strd/strn' headers found\n");
#endif
			return node;
		};

#ifdef AVIFILE_VERBOSE
		printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): buf[0] = ('%c','%c','%c','%c')\n", FOURCC_DEBUG_CHARS(buf2[0]) );
#endif

		// allocate space
		if( (*ptr = malloc(buf2[1])) == NULL)
		{
#ifdef AVIFILE_VERBOSE
        	    printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): unable to alloc %d bytes for 'node->%s'\n", buf2[1],blockname);
#endif
		    free(node);
		    return NULL;
		};

		// read block
		if( buf2[1] != fread(*ptr,1,buf2[1],file) )
		{
#ifdef AVIFILE_VERBOSE
		    printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): unable to read 'avih' block's %d bytes \n",buf2[1]);
#endif
		    free(*ptr);
		    free(node);
		    return NULL;
		};
	    
	    };    
	
	    break;
	    
	case FOURCC_CONST_HDRL:
#ifdef AVIFILE_VERBOSE
	    printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): type = FOURCC_CONST_HDRL \n");
#endif

#ifdef AVIFILE_VERBOSE
		printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): FPOS=%.8lX \n",ftell(file));
#endif
	    // load 'hdrl' header
	    if( sizeof(buf2) != fread(buf2,1,sizeof(buf2),file) )
	    {
#ifdef AVIFILE_VERBOSE
		printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): unable to read 'hdrl's header \n");
#endif
		free(node);
		return NULL;
	    };
	    
	    // check signature
	    if (FOURCC_CONST_AVIH != buf2[0])
	    {
#ifdef AVIFILE_VERBOSE
        	printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): signature != FOURCC_CONST_AVIH (signature = '%c/%c/%c/%c')\n", FOURCC_DEBUG_CHARS(buf2[0])  );
#endif
		free(node);
		return NULL;
	    };
	    
	    
	    // alloc mem for struct
	    if ( (node->avih = malloc(buf2[1])) == NULL )
	    {
#ifdef AVIFILE_VERBOSE
        	printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): unable to alloc %d bytes for 'node->avih'\n", buf2[1]);
#endif
		free(node);
		return NULL;
	    };
	    
#ifdef AVIFILE_VERBOSE
		printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): reading avih (%d bytes) at FPOS=%.8lX \n",buf2[1],ftell(file));
#endif
	    // load avih struct
	    if( buf2[1] != fread(node->avih,1,buf2[1],file) )
	    {
#ifdef AVIFILE_VERBOSE
		printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): unable to read 'avih' block's %d bytes \n",buf2[1]);
#endif
		free(node->avih);
		free(node);
		return NULL;
	    };
	    
	    break;


	case FOURCC_CONST_ODML:
#ifdef AVIFILE_VERBOSE
	    printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): type = FOURCC_CONST_ODML \n");
#endif

#ifdef AVIFILE_VERBOSE
	    printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): FPOS=%.8lX \n",ftell(file));
#endif
	    // load 'hdrl' header
	    if( sizeof(buf2) != fread(buf2,1,sizeof(buf2),file) )
	    {
#ifdef AVIFILE_VERBOSE
		printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): unable to read 'dmlh's header \n");
#endif
		free(node);
		return NULL;
	    };
	    
	    // check signature
	    if (FOURCC_CONST_DMLH != buf2[0])
	    {
#ifdef AVIFILE_VERBOSE
        	printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): signature != FOURCC_CONST_DMLH (signature = '%c/%c/%c/%c')\n", FOURCC_DEBUG_CHARS(buf2[0])  );
#endif
		free(node);
		return NULL;
	    };
	    
	    
	    // alloc mem for struct
	    if ( (node->dmlh = malloc(buf2[1])) == NULL )
	    {
#ifdef AVIFILE_VERBOSE
        	printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): unable to alloc %d bytes for 'node->dmlh'\n", buf2[1]);
#endif
		free(node);
		return NULL;
	    };
	    
#ifdef AVIFILE_VERBOSE
	    printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): reading dmlh (%d bytes) at FPOS=%.8lX \n",buf2[1],ftell(file));
#endif
	    // load avih struct
	    if( buf2[1] != fread(node->dmlh,1,buf2[1],file) )
	    {
#ifdef AVIFILE_VERBOSE
		printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): unable to read 'dmlh' block's %d bytes \n",buf2[1]);
#endif
		free(node->dmlh);
		free(node);
		return NULL;
	    };

	    return node;
	    
	    break;
    };
    
    // now search for lists
    while(1)
    {
#ifdef AVIFILE_VERBOSE
	printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): searching for LIST ... \n");
#endif

#ifdef AVIFILE_VERBOSE
	printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): SWITCH: LEVEL=%d FPOS=%.8lX \n",parse_level, ftell(file));
#endif

	// check if need to exti
#ifdef AVIFILE_VERBOSE
	printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): limits check ftell=%.8lX (%d), block_limit=%d, node->fpos=%.8X (%d), node->type=%.8X\n",ftell(file), ftell(file),block_limit, node->fpos, node->fpos, node->type);
#endif
	if((node->fpos + block_limit - sizeof(buf3) ) <=  ftell(file))
	{
	    return node;
	};

// check for file end
	if(feof(file))
	{
#ifdef AVIFILE_VERBOSE
	    printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): end of file reached \n");
#endif
	    return node;
	};
	// load 'LIST' header
	if( sizeof(buf3) != fread(buf3,1,sizeof(buf3),file) )
	{
#ifdef AVIFILE_VERBOSE
	    printf(AVIFILE_VERBOSE "ERROR(avifile_list_node_load): unable to read 'LIST' header \n");
#endif
	    free(node);
	    return NULL;
	};
	

#ifdef AVIFILE_VERBOSE
    	printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): '%c/%c/%c/%c' : %.8X  :'%c/%c/%c/%c'\n", FOURCC_DEBUG_CHARS(buf3[0]), buf3[1] , FOURCC_DEBUG_CHARS(buf3[2]) );
#endif

	last_block = NULL;

	// determinate type of block found
	switch (buf3[2])
	{
	    case FOURCC_CONST_HDRL:
		last_block = (node->hdrl = avifile_list_node_load(file,FOURCC_CONST_HDRL, parse_level + 1, buf3[1]));
		break;


	    case FOURCC_CONST_ODML:
		last_block = (node->odml = avifile_list_node_load(file,FOURCC_CONST_ODML, parse_level + 1, buf3[1]));
		break;

		
	    case FOURCC_CONST_STRL:
		if(node->strl_count)
		    node->strl = (struct avifile_list_node**)realloc(node->strl, (node->strl_count + 1) * sizeof(struct avifile_list_node*) );
		else
		    node->strl = (struct avifile_list_node**)malloc(sizeof(struct avifile_list_node*));

		last_block = (node->strl[node->strl_count] = avifile_list_node_load(file,FOURCC_CONST_STRL, parse_level + 1, buf3[1]));
		
		node->strl_count++;
		break;
		
	    case FOURCC_CONST_MOVI:
		last_block = (node->movi = avifile_list_node_load(file,FOURCC_CONST_MOVI, parse_level + 1, buf3[1]));
		fseek(file,buf3[1] - 4,SEEK_CUR);
//		fseek(file,buf3[1] - sizeof(buf3),SEEK_CUR);
#ifdef AVIFILE_VERBOSE
		printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): 'movi' section  skipped !! \n");
#endif
//		return node;
		break;

	    default:
	    {
		switch(buf3[0])
    		{
		    case FOURCC_CONST_JUNK:
#ifdef AVIFILE_VERBOSE
			printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): 'JUNK' found! \n");
#endif
			fseek(file,buf3[1] - 4,SEEK_CUR);
			break;

		    case FOURCC_CONST_IDX1:
#ifdef AVIFILE_VERBOSE
			printf(AVIFILE_VERBOSE "NOTICE(avifile_list_node_load): 'idx1' loads %d bytes\n",buf3[1]);
#endif
			fseek(file,-4,SEEK_CUR);
			node->idx1_len = buf3[1];
			node->idx1 = malloc(buf3[1]);
			fread(node->idx1,1,buf3[1],file);
			break;

		};
	    };


	};

	
    };
};

static void* avifile_parse(struct avifile_list_node* node,struct avifile_object* ao)
{
    int i,idx1_count,stream_id,idx1_offset;
    struct avifile_idx1_item* idx1;    

    // route 'avih'
    if(node->hdrl)
    {
#ifdef AVIFILE_VERBOSE
        printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): 'node->hdrl' present !\n");
#endif
        if(node->hdrl->avih)
        {
            ao->avih = node->hdrl->avih;
#ifdef AVIFILE_VERBOSE
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): 'node->hdrl->avih' present !\n");
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): ao->avih->MicroSecPerFrame=%d\n", avih_MicroSecPerFrame(ao->avih) );
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): ao->avih->TotalFrames=%d\n", avih_TotalFrames(ao->avih) );
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): ao->avih->Streams=%d\n", avih_Streams(ao->avih) );
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): ao->avih->Width=%d\n", avih_Width(ao->avih) );
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): ao->avih->Height=%d\n", avih_Height(ao->avih) );
#endif
        };
        
        if(node->hdrl->strl_count)
        {
#ifdef AVIFILE_VERBOSE
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): node->hdrl->strl_count=%d\n",node->hdrl->strl_count);
#endif
            // enum streams
            for(i=0;(i<node->hdrl->strl_count) && (i<2); i++)
            {
                if(node->hdrl->strl[i])
                {
#ifdef AVIFILE_VERBOSE
                    printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): 'node->hdrl->strl[%d]' present (strh=%.8X, strf=%.8X)!\n", i, node->hdrl->strl[i]->strh, node->hdrl->strl[i]->strf);
                    printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): node->hdrl->strl[%d]->strh.Type='%c/%c/%c/%c'\n", i, FOURCC_DEBUG_CHARS(strh_Type(node->hdrl->strl[i]->strh)) );
                    printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): node->hdrl->strl[%d]->strh.Handler='%c/%c/%c/%c'\n", i, FOURCC_DEBUG_CHARS(strh_Handler(node->hdrl->strl[i]->strh)) );
                    printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): node->hdrl->strl[%d]->strh.Length=%d'\n", i, strh_Length(node->hdrl->strl[i]->strh) );
                    printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): node->hdrl->strl[%d]->strh.SuggestedBufferSize=%d'\n", i, strh_SuggestedBufferSize(node->hdrl->strl[i]->strh) );
#endif
                    // check what is it stream
                    switch(strh_Type(node->hdrl->strl[i]->strh))
                    {
                        case FOURCC_CONST_VIDS:
                            // assign headers&formats
                            ao->video_strh = node->hdrl->strl[i]->strh;
                            ao->video_strf = node->hdrl->strl[i]->strf;
                            // assign idx1
                            ao->video_idx1_i = i;
                            break;

                        case FOURCC_CONST_AUDS:
                            // assign headers&formats
                            ao->audio_strh = node->hdrl->strl[i]->strh;
                            ao->audio_strf = node->hdrl->strl[i]->strf;
                            // assign idx1
                            ao->audio_idx1_i = i;
                            break;

                        case FOURCC_CONST_IVAS:
                            // assign headers&formats
                            ao->dv_strh = node->hdrl->strl[i]->strh;
                            ao->dv_strf = node->hdrl->strl[i]->strf;
                            ao->video_strh = node->hdrl->strl[i]->strh;
                            ao->video_strf = node->hdrl->strl[i]->strf;
                            // assign idx1
                            ao->dv_idx1_i = i;
                            break;
                        
                    };
                };
            };
            
        
        };

        // build new index table
        if(node->idx1)
        {
            idx1_count = node->idx1_len / sizeof(struct avifile_idx1_item);
            idx1 = (struct avifile_idx1_item*)node->idx1;
#ifdef AVIFILE_VERBOSE
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): 'node->idx1' present (%d bytes) !\n",node->idx1_len);
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): idx1_count=%d\n",idx1_count);
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): node->movi->fpos=%.8X (%d) \n",node->movi->fpos,node->movi->fpos);
#endif

            // try to guess: from what is counting begins: from file or from 'movi' section
            idx1_offset = (node->movi->fpos > idx1[0].ChunkOffset)?(node->movi->fpos - 4):0;

            // enumerate all indexes
            for(i = 0; i < idx1_count ; i++)
            {
                idx1[i].ChunkOffset += idx1_offset;
                stream_id = (idx1[i].ckid & 0xF00) >> 8;
//                printf("idx1[%d]: stream_id=%d, ckid=%.8X, ChunkOffset=%.8X, ChunkLength=%d \n",i, stream_id, idx1[i].ckid, idx1[i].ChunkOffset, idx1[i].ChunkLength);

                if(stream_id < AVIFILE_MAXIMUM_STREAMS)
                {
                    // check mem
                    if(ao->idx1_count[stream_id])
                        // reallocate
                        ao->idx1[stream_id] = (int*)realloc(ao->idx1[stream_id],sizeof(int)*(ao->idx1_count[stream_id] + 1));
                    else
                        // allocate news
                        ao->idx1[stream_id] = (int*)malloc(sizeof(int)*(ao->idx1_count[stream_id] + 1));

                    // save num of index
                    ao->idx1[stream_id][ao->idx1_count[stream_id]] = i;
                    ao->idx1_count[stream_id]++;
                };
            };
#ifdef AVIFILE_VERBOSE
            for(i=0;i<AVIFILE_MAXIMUM_STREAMS;i++)
                printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): ao->idx1_count[%d]=%d\n",i,ao->idx1_count[i]);
#endif

        }
        else
        {
#ifdef AVIFILE_VERBOSE
            printf(AVIFILE_VERBOSE "NOTICE(avifile_parse): 'idx1' NOT present !\n");
#endif
        };    	
    };
    
    return NULL;
};

struct avifile_object* avifile_load(char* filename)
{
    struct avifile_list_node* node;
    struct avifile_object* ao = (struct avifile_object*)malloc(sizeof(struct avifile_object));
    unsigned int buf[3];
    
    CLEAR(*ao);
    
    if( (ao->file = fopen(filename, "rb")) == NULL)
    {
#ifdef AVIFILE_VERBOSE
        printf(AVIFILE_VERBOSE "ERROR(avifile_load): unable to load file '%s'\n",filename);
#endif
        free(ao);
        return NULL;
    };
    
    // load 'LIST' header
    if( sizeof(buf) != fread(buf,1,sizeof(buf),(FILE*)ao->file) )
    {
#ifdef AVIFILE_VERBOSE
        printf(AVIFILE_VERBOSE "ERROR(avifile_load): unable to read 'RIFF' header \n");
#endif
        fclose((FILE*)ao->file);
        free(ao);
        return NULL;
    };
    
    // check if real RIFF header
    if
    (
	(buf[0] == FOURCC_CONST_RIFF)
	&&
	(buf[2] == FOURCC_CONST_AVI)
    )
    {
#ifdef AVIFILE_VERBOSE
	printf(AVIFILE_VERBOSE "NOTICE(avifile_load): RIFF AVI header found !\n");
#endif
	if((node = avifile_list_node_load((FILE*)ao->file,FOURCC_CONST_RIFF,0,buf[1])) != NULL)
	{
	    // a bit of parse
	    avifile_parse(node,ao);
	
	    // assign riff tree
            ao->riff = node;
            
	    // chech for compleated
	    if((ao->avih) && (ao->video_strh))
	        return ao;
	};

    };
    
#ifdef AVIFILE_VERBOSE
    printf(AVIFILE_VERBOSE "ERROR(avifile_load): RIFF AVI header not found \n");
#endif

    fclose((FILE*)ao->file);
    free(ao);
    
    return NULL;
};


