#ifndef _AVIFILE_H
#define _AVIFILE_H

#define FOURCC_DIRECT(C1,C2,C3,C4) (  (((unsigned int)((unsigned char)C1)) << 0) | (((unsigned int)((unsigned char)C2)) << 8) | (((unsigned int)((unsigned char)C3)) << 16) | (((unsigned int)((unsigned char)C4)) << 24) )

// avih access macros
#define avih_MicroSecPerFrame(PTR)	(*((int*)(((unsigned char*)PTR) +  0)))
#define avih_MaxBytesPerSec(PTR)	(*((int*)(((unsigned char*)PTR) +  4)))
#define avih_Reserved1(PTR)		(*((int*)(((unsigned char*)PTR) +  8)))
#define avih_Flags(PTR)			(*((int*)(((unsigned char*)PTR) + 12)))
#define avih_TotalFrames(PTR)		(*((int*)(((unsigned char*)PTR) + 16)))
#define avih_InitialFrames(PTR)		(*((int*)(((unsigned char*)PTR) + 20)))
#define avih_Streams(PTR)		(*((int*)(((unsigned char*)PTR) + 24)))
#define avih_SuggestedBufferSize(PTR)	(*((int*)(((unsigned char*)PTR) + 28)))
#define avih_Width(PTR)			(*((int*)(((unsigned char*)PTR) + 32)))
#define avih_Height(PTR)		(*((int*)(((unsigned char*)PTR) + 36)))
#define avih_Reserved(PTR)		((int*)(((unsigned char*)PTR) + 40))

// strh access macros
#define strh_Type(PTR)			(*((int*)(((unsigned char*)PTR) +  0)))
#define strh_Handler(PTR)		(*((int*)(((unsigned char*)PTR) +  4)))
#define strh_Flags(PTR)			(*((int*)(((unsigned char*)PTR) +  8)))
#define strh_Priority(PTR)		(*((int*)(((unsigned char*)PTR) + 12)))
#define strh_InitialFrames(PTR)		(*((int*)(((unsigned char*)PTR) + 16)))
#define strh_Scale(PTR)			(*((int*)(((unsigned char*)PTR) + 20)))
#define strh_Rate(PTR)			(*((int*)(((unsigned char*)PTR) + 24)))
#define strh_Start(PTR)			(*((int*)(((unsigned char*)PTR) + 28)))
#define strh_Length(PTR)		(*((int*)(((unsigned char*)PTR) + 32)))
#define strh_SuggestedBufferSize(PTR)	(*((int*)(((unsigned char*)PTR) + 36)))
#define strh_Quality(PTR)		(*((int*)(((unsigned char*)PTR) + 40)))
#define strh_SampleSize(PTR)		(*((int*)(((unsigned char*)PTR) + 44)))
#define strh_Frame(PTR)			((short int*)(((unsigned char*)PTR) +  48))

// strf_video access macros
#define strf_video_Size(PTR)		(*((int*)(((unsigned char*)PTR) +  0)))
#define strf_video_Width(PTR)		(*((int*)(((unsigned char*)PTR) +  4)))
#define strf_video_Height(PTR)		(*((int*)(((unsigned char*)PTR) +  8)))
#define strf_video_Planes(PTR)		(*((short int*)(((unsigned char*)PTR) + 12)))
#define strf_video_BitCount(PTR)	(*((short int*)(((unsigned char*)PTR) + 14)))
#define strf_video_Compression(PTR)	(*((int*)(((unsigned char*)PTR) + 16)))
#define strf_video_SizeImage(PTR)	(*((int*)(((unsigned char*)PTR) + 20)))
#define strf_video_XPelsPerMeter(PTR)	(*((int*)(((unsigned char*)PTR) + 24)))
#define strf_video_YPelsPerMeter(PTR)	(*((int*)(((unsigned char*)PTR) + 28)))
#define strf_video_ClrUsed(PTR)		(*((int*)(((unsigned char*)PTR) + 32)))
#define strf_video_ClrImportant(PTR)	(*((int*)(((unsigned char*)PTR) + 36)))

// strf_dv access macros
#define strf_dv_AuxSrc(PTR)		(*((int*)(((unsigned char*)PTR) +  0)))
#define strf_dv_AAuxCtl(PTR)		(*((int*)(((unsigned char*)PTR) +  4)))
#define strf_dv_AAuxSrc1(PTR)		(*((int*)(((unsigned char*)PTR) +  8)))
#define strf_dv_AAuxCtl1(PTR)		(*((int*)(((unsigned char*)PTR) + 16)))
#define strf_dv_VAuxSrc(PTR)		(*((int*)(((unsigned char*)PTR) + 24)))
#define strf_dv_VAuxCtl(PTR)		(*((int*)(((unsigned char*)PTR) + 32)))
#define strf_dv_Reserved(PTR)		(((int*)(((unsigned char*)PTR) +  36)))

#define strf_audio_FormatTag(PTR)	(*((short int*)(((unsigned char*)PTR) +  0)))		/* format type */
#define strf_audio_Channels(PTR)	(*((short int*)(((unsigned char*)PTR) +  2)))		/* number of channels (i.e. mono, stereo...) */
#define strf_audio_SamplesPerSec(PTR)	(*((int*)(((unsigned char*)PTR) +  4)))		/* sample rate */
#define strf_audio_AvgBytesPerSec(PTR)	(*((int*)(((unsigned char*)PTR) +  8)))		/* for buffer estimation */
#define strf_audio_BlockAlign(PTR)	(*((short int*)(((unsigned char*)PTR) +  12)))	/* block size of data */
#define strf_audio_BitsPerSample(PTR)	(*((short int*)(((unsigned char*)PTR) +  14)))	/* Number of bits per sample of mono data */
#define strf_audio_Size(PTR)		(*((short int*)(((unsigned char*)PTR) +  16)))	/* The count in bytes of the extra size */

#define AVIFILE_MAXIMUM_STREAMS 2

struct avifile_object
{
    // struct avifile_list_node*
    void* riff;
    int* idx1[AVIFILE_MAXIMUM_STREAMS];
    int  idx1_count[AVIFILE_MAXIMUM_STREAMS];
    
    // FILE* file;
    void* file;

    void* avih;    

    void* video_strh;
    void* video_strf;
    int  video_idx1_i;

    void* audio_strh;
    void* audio_strf;
    int  audio_idx1_i;

    void* dv_strh;
    void* dv_strf;
    int  dv_idx1_i;

};

struct avifile_subchunk_item
{
    long chunk_i;
    long offset;
};
	


// avi file load&parse
struct avifile_object* avifile_load(char* filename);
void avifile_close(struct avifile_object* node);

// load/free chunks
void* avifile_chunk_load(struct avifile_object* ao,int stream,int num);
void avifile_chunk_free(void* chunk);
int avifile_chunk_length(struct avifile_object* ao, int stream, int num);

// subchunk index table build/free
struct avifile_subchunk_item* avifile_subchunks_index_build(struct avifile_object* ao, int subchunk_size, long subchunk_count, void* idx, int idx_len);
void avifile_subchunks_index_free(struct avifile_subchunk_item* p);


#endif	// _AVIFILE_H

