#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "frames_source.h"
#include "frames_source_avi.h"
#include "frames_source_raw.h"

char frames_source_path[1024];

//#define DEBUG

frames_source::frames_source(char* filename)
{
    // reset flags
    _file_handler = NULL;
    _file_name = NULL;
    _operation_ready = 0;
    _frames_count = 0;
    
    // alloc space for filename
    _file_name = (char*)malloc(strlen(filename) + 1);
    strcpy((char*)_file_name, filename);
    
    // try to open file
    _file_handler = fopen(_file_name, "rb");
};

frames_source::~frames_source()
{
    // release file name
    free(_file_name);
    
    // if file is open - close it
    if(_file_handler)
	fclose((FILE*)_file_handler);
};
    
int frames_source::ready()
{
    return (_frames_count > 0)?1:0;
};

int frames_source::frames()
{
    return _frames_count;
};


/*
    This function tries guess files
*/    

static char* open_frame_source_exts[] = {".avi",".raw",NULL};

frames_source* open_frame_source(char* id)
{
    char filename[1024];
    frames_source* src = NULL;
    
    // enum supported formats and try to guess
    for(int i=0;open_frame_source_exts[i];i++)
    {

	// compose full file name
	sprintf(filename,"%s/%s%s",frames_source_path,id,open_frame_source_exts[i]);

#ifdef DEBUG
printf("opening file: '%s'\n",filename);
#endif

	
	// each number - each class
	switch(i)
	{
	    case 0: src = new frames_source_avi(filename);	break;
	    case 1: src = new frames_source_raw(filename);	break;
	    default:
		return NULL;
	};

#ifdef DEBUG
printf("ready=%d\n",src->ready());
#endif
	
	// test if all OK
	if(src->ready())
	    return src;
	else
	{
	    // in a case of fail free object
	    delete src;
	    src = NULL;
	};
    };

    // nothing found - return    
    return NULL;
};

