#include <stdlib.h>
#include <string.h>

#include "frames_source.h"
#include "frames_source_avi.h"

// avi file operations
#include "avifile.h"

// lib dv operations                                                                                                                                               
#include "libdv/dv.h"
#include "libdv/enc_audio_input.h"
#include "libdv/enc_output.h"

#define CLEAR(x) memset (&(x), 0, sizeof (x))

#define FOURCC_DVSD FOURCC_DIRECT('d','v','s','d')

static void convert_le16(void* _src,int bytes)
{
    unsigned char *src = (unsigned char*)_src;
    unsigned char temp;
    for(int i=0;i<bytes;i+=2,src += 2)
    {
        temp = *(src + 1); *(src + 1) = *src; *src = temp;
    };
};

static void convert_le16(void* _dst,void* _src,int bytes)
{
    unsigned char *src = (unsigned char*)_src;
    unsigned char *dst = (unsigned char*)_dst;
    for(int i=0;i<bytes;i+=2,src += 2, dst+=2)
    {
        *(dst + 1) = *src;
        *dst = *(src + 1);
    };
};
                                    



#define AO ((struct avifile_object*)_ao)

frames_source_avi::frames_source_avi(char* filename) : frames_source(filename)
{
    _dv_type = 0;
    _audio = NULL;
    _ao = NULL;
    _audio_chunk_index = NULL;
    _raw_audio = NULL;
    _raw_video = NULL;

    // ok, let try to open file
    if(_file_handler)
        _ao  = (struct avifile_object*)avifile_load(_file_name);
    
    // check if it real avi
    if(_ao)
    {
	// check DV type and audio presence
	if(AO->dv_strh)
	{
	    // seems video is native DV
	    _dv_type = 1;
	}
	else
	{
	    // check if video present :-))
	    if(AO->video_strh)
	    {
		// check if handler is correct
		if( strh_Handler(AO->video_strh) == FOURCC_DVSD )
		{
		    // ok given video is type 2
		    _dv_type = 2;
		    
		    // check if audio present
		    if(AO->audio_strh)
		    {
			// allocate space for audio struct
			_audio = malloc(sizeof(dv_enc_audio_info_t));
			CLEAR(*((dv_enc_audio_info_t*)_audio));

			// fill struct
			((dv_enc_audio_info_t*)_audio)->channels	= (int)strf_audio_Channels(AO->audio_strf);
	    		((dv_enc_audio_info_t*)_audio)->frequency	= (int)strf_audio_SamplesPerSec(AO->audio_strf);
		        ((dv_enc_audio_info_t*)_audio)->bitspersample	= (int)strf_audio_BitsPerSample(AO->audio_strf);
			((dv_enc_audio_info_t*)_audio)->bytealignment	= (int)strf_audio_BlockAlign(AO->audio_strf);
			((dv_enc_audio_info_t*)_audio)->bytespersecond	= strf_audio_AvgBytesPerSec(AO->audio_strf);
			((dv_enc_audio_info_t*)_audio)->bytesperframe	= strf_audio_AvgBytesPerSec(AO->audio_strf)/(1000000 / avih_MicroSecPerFrame(AO->avih));
			
            		// build audio chunks additional index
            		_audio_chunk_index = avifile_subchunks_index_build
            		(
			    AO,
            		    ((dv_enc_audio_info_t*)_audio)->bytesperframe,
            		    avih_TotalFrames(AO->avih),
            		    AO->idx1[AO->audio_idx1_i],
                            AO->idx1_count[AO->audio_idx1_i]
                        );
/*			
			{
			    FILE* t0 = fopen("_audio_chunk_index","wt");
			    for(int i=0;i<avih_TotalFrames(AO->avih);i++)
				fprintf
				(
				    t0,
				    "_audio_chunk_index[%3d] = {chunk_i=%4d, offset=%4d}\n",
				    i,
				    ((struct avifile_subchunk_item*)_audio_chunk_index)[i].chunk_i,
				    ((struct avifile_subchunk_item*)_audio_chunk_index)[i].offset
				);
			    fclose(t0);
			};
*/
		    }
		};
	    };
	};
	
	// load frames count
	if (_dv_type > 0)
	{
	    _frames_count = avih_TotalFrames(AO->avih);
	    _video_chunk_i = 0xFFFFFFFF;
	    _audio_chunk_i = 0xFFFFFFFF;
	};
    };
    
    
};


frames_source_avi::~frames_source_avi()
{

    if (_audio)
	free(_audio);
	
    if (_ao)
	avifile_close(AO);

    if(_raw_video)
        avifile_chunk_free(_raw_video);

    if(_raw_audio)
        avifile_chunk_free(_raw_audio);
        
    if (_audio_chunk_index)
        avifile_subchunks_index_free((struct avifile_subchunk_item*)_audio_chunk_index);
    
};

char* frames_source_avi::whoami()
{
    return "frame_source_avi";
};


int frames_source_avi::load(void* buffer, int frame_number)
{
    // first check if given number in range
    if((frame_number >= 0) && (_frames_count > 0) && (frame_number < _frames_count))
    {
        // calculate video chunk
        int video_chunk_i = frame_number;

        // determinate if current frame should reloaded
        if(video_chunk_i != _video_chunk_i)
        {
            // should be loaded!
            
            // check if previous video chunk should be freed
            if(_raw_video)
                avifile_chunk_free(_raw_video);
                
            // load new video chunk
            _raw_video = avifile_chunk_load(AO,AO->video_idx1_i,(_video_chunk_i = video_chunk_i) );
	    
	    // if video frame is not loaded - return
	    if(!(_raw_video))
		return 0;
        }
        
        // check if we need to load audio
        if(_audio)
        {
            // check if we need to reload chunk
            if( _audio_chunk_i != ((struct avifile_subchunk_item*)_audio_chunk_index)[frame_number].chunk_i )
            {
                // check if we need to free audio chunk
                if(_raw_audio)
                    avifile_chunk_free(_raw_audio);
                    
                // load new audio chunk
                _raw_audio = avifile_chunk_load
		(
		    AO,
		    AO->audio_idx1_i, 
		    ( _audio_chunk_i = ((struct avifile_subchunk_item*)_audio_chunk_index)[frame_number].chunk_i ) 
		);
            };
            
        };

        // fill output buffer
        memcpy(buffer, _raw_video, 144000);
        
        // embed audio using libdv
        if(_audio)
        {
	    // fill and encode buffer of audio struct
            convert_le16
            (
	        ((dv_enc_audio_info_t*)_audio)->data,
	        (void*)
		(
		    (int)_raw_audio
		    +
		    ((dv_enc_audio_info_t*)_audio)->bytesperframe
		    *
		    ((struct avifile_subchunk_item*)_audio_chunk_index)[frame_number].offset  
		),
		((dv_enc_audio_info_t*)_audio)->bytesperframe
	    );
													    
            _dv_raw_insert_audio((unsigned char*)buffer, (dv_enc_audio_info_t*)_audio, 1);
        };

/*
	{
	    char t0[1024];sprintf(t0,"audio/%.4d_audio.raw",frame_number);
	    FILE* t1 = fopen(t0,"wb");
	    fwrite
	    (
	        (void*)
		(
		    (int)_raw_audio
		    +
		    ((dv_enc_audio_info_t*)_audio)->bytesperframe
		    *
		    ((struct avifile_subchunk_item*)_audio_chunk_index)[frame_number].offset  
		),
		1,
		((dv_enc_audio_info_t*)_audio)->bytesperframe,
		t1
	    );
	    fclose(t1);
	};
*/

	return 1;

    };
    
    return 0;
};

