/*

Changelog:
    2005-11-11:
	*Threads priority adjastment.

    2005-11-07:
	*'new_cue' code rewriten - staight and simple, not hooks.
	*Problem found with first frames is not 0 , it is 1 ! FIXED !
	
	*To provide loop functionlity add 'playback_remains[]' array for storing 
	info about scheduled length of items

    2005-11-04 04:00:00
	*Seems model is working !!! Let's test and fix!
	
*/


#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <time.h>

#include "schd_replacer.h"
#include "frames_source.h"
#include "black_dv_frame.h"

//#define LOG_FRAME_REQUEST
//#define LOG_FIRST_FRAMES

#define DV_PAL_FRAME_SIZE 144000
static void** buffers;
static long buffer_length;
static long first_clear_frame;
static long active_frame;
static long clear_frames;
static pthread_mutex_t lock_buffer_state;
static long new_cue;

static frames_source** sources;
static long sources_count;
static pthread_mutex_t lock_sources;

static long global2local(struct instance_item* instance, long global_frame_no, long *source_index,long *frame_no, char* caller_id)
{
    long i;
    long current_position = 0;
    
    for(i=0; i< sources_count; i++)
    {
	// chech if source present
	if(sources[i])
	{
	    // detect if given global frame is in range is i-th item
	    if
	    (
		(current_position <= global_frame_no )
		&&
		(global_frame_no < (current_position + instance->current_block->events[i].dur) )
	    )
	    {
		// YES ! given global is inside of i-th id
		
		// assign source index
		*source_index = i;
		
		// calc frame no inside of reel
		*frame_no = instance->current_block->events[i].tc + (global_frame_no - current_position);
		
		return 1;
	    };
	    current_position += instance->current_block->events[i].dur;
	};
    };    

    // hm? something wrong - OK lets global frame will be last frame in last source
    syslog
    (
	LOG_NOTICE, 
	"global2local: something wrong. global_frame_no(%ld) is out of range (caller='%s',playback_length=%ld,playback_remain=%ld,cacheman_frames_sent=%ld,cacheman_frames_loaded=%ld,clear_frames=%ld", 
	global_frame_no, 
	caller_id, 
	instance->playback_length, 
	instance->playback_remain,
	instance->cacheman_frames_sent,
	instance->cacheman_frames_loaded,
	clear_frames
    );
    if(global2local(instance, current_position - 1, source_index, frame_no, "global2local()"))
	return 0;
	
    syslog(LOG_NOTICE, "global2local: problem!!!!");	
    
    return 0;
};


// return current active frame
void* get_active_frame(struct instance_item* instance, long* index)
{
#ifdef LOG_FRAME_REQUEST
    struct timeval current_time;
    gettimeofday(&current_time, NULL);
    syslog
    (
	LOG_NOTICE,
	"get_active_frame(): active_frame=%d,playback_remain=%d,cacheman_frames_sent=%d,playback_length=%d  at %ld.%.6ld",
	active_frame,
	instance->playback_remain, 
	instance->cacheman_frames_sent , 
	instance->playback_length,
	current_time.tv_sec,current_time.tv_usec
    );
#endif

    return buffers[*index = active_frame];
};

void mark_active_frame_clear(struct instance_item* instance)
{
    // we increment active frame in a case of 
    // playback state and no more frames should
    // be played according to status
    if
    (
	(instance->state == SCHD_PLAYBACK)		// normal playback condition
	&&
	(instance->playback_remain > 0)
    )
    {
#ifdef LOG_FIRST_FRAMES
	if((instance->playback_length - instance->playback_remain) < 30)
	{
    	    struct timeval current_time;
	    gettimeofday(&current_time, NULL);
	    syslog(LOG_NOTICE, "mark_active_frame_clear(): playback_remain=%d,cacheman_frames_sent=%d,playback_length=%d  at %ld.%.6ld",
		instance->playback_remain, instance->cacheman_frames_sent , instance->playback_length,
		current_time.tv_sec,current_time.tv_usec);
	};
#endif
    
    	// lock
	pthread_mutex_lock(&lock_buffer_state);
	
	// underflow?
	if((clear_frames + 1) >= buffer_length)
	{
	    // UNDERFLOW !!!!
	    syslog(LOG_NOTICE, "mark_active_frame_clear(): buffer underrun %d frames!", buffer_length - clear_frames );
	}
	else
	{
	    // increment number of clear frames
	    clear_frames++;
	    
	    // increment frames sent counter
	    instance->cacheman_frames_sent++;

	    // decrement number of playback frames left
	    if((instance->playback_remain == 1) && (instance->f_loop_block))
	    {
		// reset 
		instance->playback_remain = instance->playback_length;
	    
		// reset remains
		for(int i=0;i<instance->playback_count;i++)
		    instance->playback_remains[i] = instance->playback_lengths[i];

		// notice	
		syslog(LOG_NOTICE, "mark_active_frame_clear(): loop active - resetting 'playback_remain'");
	    }
	    else
	    {
		// no loop
		instance->playback_remain--;
	    
		// decrement remains
		long source_index,frame_no;
		if(global2local(instance,instance->playback_length - instance->playback_remain - 1, &source_index, &frame_no, "mark_active_frame_clear()"))
		    instance->playback_remains[source_index] -- ;
	    };
	    
	    // first active frame moves forward
	    active_frame = ( (active_frame + 1) % buffer_length );
	    
	};
	
    	// unlock
	pthread_mutex_unlock(&lock_buffer_state);
    };
};




static void cache(struct instance_item* instance)
{
    // lock source
    pthread_mutex_lock(&lock_sources);

    if(new_cue)
    {
    	// lock
	pthread_mutex_lock(&lock_buffer_state);

	// reset state
	clear_frames = buffer_length;
	first_clear_frame = 0;
	instance->cacheman_frames_loaded = 0;
	instance->cacheman_frames_sent = 0;
	active_frame = 0;
	
    	// unlock
	pthread_mutex_unlock(&lock_buffer_state);
	
	// reset trigger
	new_cue = 0;
    };


    pthread_mutex_lock(&lock_buffer_state);
    long _clear_frames = clear_frames;
    pthread_mutex_unlock(&lock_buffer_state);


    // check if clear frams present
    if( (_clear_frames)&&(instance->cacheman_frames_loaded <= instance->playback_length) )
    {
	long frames_loaded = 0;

	// load clear frames
	for(long i=0,source_index,frame_no; i< _clear_frames ; i++)
	{
	    // calc global frame no to load
	    long global_frame_no = i + instance->cacheman_frames_loaded;

	    // calc source id and frame inside of source
	    global2local(instance,global_frame_no, &source_index, &frame_no, "cache()");
	    
	    // use source method to load frame
	    if(!(sources[source_index]->load( buffers[first_clear_frame], frame_no) ))
		syslog(LOG_NOTICE, "cache(): failed to load frame! global=%ld,source=%ld,local=%ld", global_frame_no, source_index, frame_no);

	    // move position of first active frame forward
	    first_clear_frame = ( (first_clear_frame + 1) % buffer_length );
	    
	    frames_loaded++;
	    
	    // make a little sleep for scheduler
	    struct timespec s = {0,1};
	    nanosleep(&s,NULL);
	};

	instance->cacheman_frames_loaded += frames_loaded;

    	// lock
	pthread_mutex_lock(&lock_buffer_state);
	// decrement clear frames count
	clear_frames -= frames_loaded;
    	// unlock
	pthread_mutex_unlock(&lock_buffer_state);
	
    }
    else
    {
	//chech if loaded frames reached and and loop mode is on?
	if
	(
	    (instance->cacheman_frames_loaded > instance->playback_length)	// pointer over 
	    &&
	    (instance->f_loop_block)
	)
	{
	    // loop active and loaded frames pointer reached end
	    
	    // reset it to zero !
	    instance->cacheman_frames_loaded = 0;
	    syslog(LOG_NOTICE, "cache(): loop active - resetting 'cacheman_frames_loaded'");
	};
    
    };

    // unlock sources
    pthread_mutex_unlock(&lock_sources);

};

static long cue(struct instance_item* instance)
{
    long result = 1;

    // lock source
    pthread_mutex_lock(&lock_sources);
    
    // lock block
    pthread_mutex_lock(&instance->lock_block_update);

#ifdef CACHEMAN_VERBOSE
//syslog(LOG_NOTICE, "cue: ", id, dur);
#endif
    
    // check block
    if(instance->current_block)
    {
        if (instance->current_block->events_count)
        {
            // events present !!!!
            
            // reallocate space for events
            if(instance->current_block->events_count > sources_count)
            {
            	sources = (frames_source**) realloc(sources, sizeof(frames_source*) * instance->current_block->events_count );
            	instance->playback_remains = (long*) realloc(instance->playback_remains, sizeof(long*) * instance->current_block->events_count );
            	instance->playback_lengths = (long*) realloc(instance->playback_lengths, sizeof(long*) * instance->current_block->events_count );
	    };

            // assign new lengths
            instance->playback_count = (sources_count = instance->current_block->events_count);
            
            // load all item
            instance->playback_remain = 0;
            instance->playback_length = 0;
            for(int i=0; i<sources_count; i++)
            {
            	// store in local vars for shorter paths
                schd_event_item* event = &instance->current_block->events[i];
                char *id = event->id ;
                
		// make a little sleep for scheduler
		struct timespec s = {0,1};
		nanosleep(&s,NULL);
		
		
                if(sources[i] = open_frame_source( id ))
                {
		    long dur = sources[i]->frames();
		    
		    // check if given range is in avi dur
		    if(dur >= (event->dur + event->tc))
		    {
                    	instance->playback_lengths[i] = (instance->playback_remains[i] = event->dur);
                    	instance->playback_remain += event->dur;
                    	instance->playback_length += event->dur;

                    	syslog(LOG_NOTICE, "cue: '%s' OK[%ld]", id, dur);
		    }
		    else
		    {
		    	syslog(LOG_NOTICE, "cue: '%s' FAIL! total frames=%ld, asked tc=%ld, dur=%ld is over length", id, dur, event->tc,event->dur);
		    	// this item is bad?
		    	if(dur > event->tc)
		    	{
			    // tc in range !!! - trim this item
			    event->dur = dur - event->tc;

                    	    instance->playback_lengths[i] = (instance->playback_remains[i] = event->dur);
                    	    instance->playback_remain += event->dur;
                    	    instance->playback_length += event->dur;
                    	    
                    	    // notify event trimmed
                    	    syslog(LOG_NOTICE, "cue: event with id '%s' trimmed to dur [%ld]", id, event->dur);
		    	}
		    	else
		    	{
		    	    instance->playback_lengths[i] = (instance->playback_remains[i] = 0);
		    	    delete sources[i];
		    	    sources[i] = NULL;
		    	};
		    };
                }
                else
                {
		    instance->playback_lengths[i] = (instance->playback_remains[i] = 0);
		    syslog(LOG_NOTICE, "cue: '%s' FAILED!", id);
                };
            };
	    
	    // check if reel present? 
	    if(!(instance->playback_length))
	    {
		syslog(LOG_NOTICE, "cue: FAIL! empty block");
		result = 0;
	    }
	    else
	    {
        	new_cue = 1;							// setup flag to jump active frame forward
        	result = 1;
	    };
        }
        else
	{
            result = 0;
	    syslog(LOG_NOTICE, "cue: Fail! No 'events; in 'current_block'");
	};
    }
    else
    {
        syslog(LOG_NOTICE, "cue: Fail! No 'current_block'");
    	result = 0;
    };

    // unlock sources
    pthread_mutex_unlock(&lock_sources);

    // unlock block if cue fails
    if(!(result))
	pthread_mutex_unlock(&instance->lock_block_update);
    
    return result;
};

static long decue(struct instance_item* instance)
{
    int i;
    
    // lock source
    pthread_mutex_lock(&lock_sources);
    
    // free all items and reset length
    for(i = 0; i < sources_count ; i++)
        if(sources[i])
        {
    	    delete sources[i];
    	    sources[i] = 0;
    	};
    instance->playback_count = (sources_count = 0);
    instance->playback_remain = 0;
    instance->playback_length = 0;

    // unlock sources
    pthread_mutex_unlock(&lock_sources);
    
    // unlock block
    pthread_mutex_unlock(&instance->lock_block_update);
    return 1;
};


// cacheman loop
static void* loop_cacheman(void* p)
{
    struct timespec sleep_time; sleep_time.tv_sec = 0; sleep_time.tv_nsec =  1000;
    struct timespec slept_time;
    struct timeval current_time;

    // cast instance ctrl struct
    struct instance_item* instance = (struct instance_item*)p;

    // setup flag thread active
    instance->f_cacheman = 1;
    syslog(LOG_NOTICE, "loop_cacheman: started");

    while((instance->f_cacheman)&&(!(instance->f_exit)))
    {
        // get current time
        gettimeofday(&current_time,NULL);
                 
	// do job depends on sys state    	
    	switch(instance->state)
    	{
	    case SCHD_PLAYBACK:
	    case SCHD_CUE_DONE:
	    	// we do only load buffers proc
	    	cache(instance);
	    	break;
	    	
	    case SCHD_CUE_INIT:
		// check if we allready cued
		if( (instance->f_cue_init) && (!(instance->f_cue_done)) )
		{
		    if(cue(instance))
		    {
		    	// cue was successfull
		    	instance->f_cue_done = 1;
		    	syslog(LOG_NOTICE, "loop_cacheman: cue() OK at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
		    }
		    else
		    {
		    	// cue was not successfull
		    	syslog(LOG_NOTICE, "loop_cacheman: cue() FAIL at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
		    	instance->f_cue_init = 0;
		    };
		};
		break;

	    case SCHD_POSTROLL:
	    	// on this destroy frames objects
	    	if (instance->f_cue_done)
	    	{
		    decue(instance);
		    instance->f_cue_done = 0;
		    syslog(LOG_NOTICE, "loop_cacheman: decue() at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
	    	};
	    	break;
	};

	// sleep a bit	
    	nanosleep(&sleep_time,&slept_time);
    };

    // exit flag happen - unlock mutex
    if (instance->f_cue_done)
    	decue(instance);
    	
    
    // setup flag thread active
    instance->f_cacheman = 0;

    // return value
    syslog(LOG_NOTICE, "loop_cacheman: finished");
    pthread_exit(0);
    
    return NULL;
};

int stop_cacheman(struct instance_item* instance)
{
    int i;
    
    // destroy buffers
    pthread_mutex_destroy(&lock_sources);
    pthread_mutex_destroy(&lock_buffer_state);

    void* loop_cacheman_exit;
    
    if(instance->f_cacheman)
    {
    	syslog(LOG_NOTICE, "stop_cacheman: Waiting for 'cacheman' thread exit");
    	instance->f_cacheman = 0;
    	if(pthread_join(instance->th_cacheman, &loop_cacheman_exit) < 0)
	    syslog(LOG_NOTICE, "stop_cacheman: Something wrong with joining to 'cacheman' thread");
	else
	    syslog(LOG_NOTICE, "stop_cacheman: 'cacheman' exits with '0x%.8X' ", (long)loop_cacheman_exit);
    }
    else
	syslog(LOG_NOTICE, "stop_cacheman: 'cacheman' seems allready finished or not runned");

    // free all buffers then     
    for(i=0;i<buffer_length;i++)
    	free(buffers[i]);
    free(buffers);

    // destroy frames sourcs
    for(i=0; i< sources_count; i++)
	delete sources[i];
    free(sources);
    sources_count = 0;
    
    
    // no thread started - exit
    return 1;
    
    
};

int start_cacheman(struct instance_item* instance)
{
    // init framesource path
    strcpy(frames_source_path,instance->cfg_cacheman_video_path);

    // init buffers
    buffer_length = instance->cfg_cacheman_buffer_length;
    buffers = (void**)malloc(sizeof(void*)*buffer_length);
    for(int i=0;i<buffer_length;i++)
    {
    	// allocate
    	buffers[i] = (void*)malloc(DV_PAL_FRAME_SIZE);
    	
    	// copy black frame
    	memcpy(buffers[i],black_dv_frame, DV_PAL_FRAME_SIZE);
    };
    first_clear_frame = -1;
    active_frame = 0;
    clear_frames = 0;
    new_cue = 0;
    pthread_mutex_init(&lock_buffer_state, NULL);
    pthread_mutex_init(&lock_sources, NULL);

    // init frames source
    sources = (frames_source**)malloc(0);
    sources_count = 0;
    instance->playback_remains = (long*)malloc(0);
    instance->playback_lengths = (long*)malloc(0);
    

    if(pthread_create(&instance->th_cacheman, &instance->th_cacheman_attr, loop_cacheman, instance) < 0)
    {
	syslog(LOG_ERR, "start_cacheman: Unable to start 'cacheman' thread!");
	return 0;
    }
    else
	return 1;
};
