/*

Changelog:
    2005-11-11:
	*Threads priority adjastment.

    2005-11-09:
	*Auxilarity blocks file 
	*Block sys_id displaying added
	
    2005-11-08:
	*Instance number displaying added.
	
    2005-11-07:
	*Loop block precessing.
    
    2005-11-06:
	*Enabling controls over extcmd and switcher operations.
	*Block length display real length in CUE_DONE and PLAYBACK modes	

    2005-11-05:
	*Force cue operation by added - rising 'f_force_cue' flag

    2005-11-04 04:00:00:
    
	*Seems model is working !!! Let's test and fix!
	
*/

#include <locale.h>
//#include <langinfo.h>
#include <ncursesw/ncurses.h>
#include <unistd.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <time.h>

#include "schd_replacer.h"
#include "mytimecode.h"
#include "statectl.h"

static char* states_str[] = 
{
    "IDLE",
    "CUE INIT",
    "CUE DONE",
    "PLAYBACK",
    "POSTROLL"
};


static void timeval2tc(timeval t, char* buf)
{
    struct tm temp_tm;
    char temp_str[128];

    // convert secs
    localtime_r(&t.tv_sec,&temp_tm);
    
    // format datetime
    strftime(buf,100,STRF_DATETIME,&temp_tm);
    
    // append frames
    sprintf(temp_str, ":%.2d", t.tv_usec/40000 ); strcat(buf, temp_str);
};

static char current_time_str[1024];
static char remain_time_str[1024] = "";


// screen update loop
static void* loop_display(void* p)
{
    int i;
    
    // cast instance ctrl struct
    struct instance_item* instance = (struct instance_item*)p;

    // refresh setup
    long refresh_counter = instance->cfg_display_refresh_interval;
    struct timespec sleep_time; sleep_time.tv_sec = 0; sleep_time.tv_nsec =  40*1000;
    struct timespec slept_time;
    struct timeval current_time;


    // setup flag thread active
    instance->f_display = 1;
    syslog(LOG_NOTICE, "loop_display: started");

    
    while((instance->f_display)&&(!(instance->f_exit)))
    {
	// sleep a bit
	nanosleep(&sleep_time,&slept_time);
	
	// update refresh counter
	refresh_counter++;

	// check if time elapsed
	if(refresh_counter > instance->cfg_display_refresh_interval)
	{
	    // retrive info about screen
	    int rows,cols;
	    getmaxyx(stdscr, rows, cols);
	
	    // reset counter
	    refresh_counter = 0;
	    
	    // current time
	    gettimeofday(&current_time, NULL);
	    timeval2tc(current_time,current_time_str);
	    
	    // calc time remains
	    pthread_mutex_lock(&instance->lock_block_read);
	    double remain = 0.0;
	    if( instance->state == SCHD_IDLE )
	    {
		// system in idle state - idle is time to
		// CUE/INIT state if current present
		
		if(instance->current_block == NULL)
		    // no current block !
    		    remain_time_str[0] = 0;
		else
		    tc_sec2txt
		    (
			(double) (instance->current_block->planned_from.tv_sec - current_time.tv_sec)
			+
			( (double) (instance->current_block->planned_from.tv_usec - current_time.tv_usec) ) / (double)1000000.0,
			remain_time_str,
			1
		    );
	    }
	    else if ( instance->state == SCHD_CUE_DONE )
	    {
		// cue done -  remains is time to WAIT TILL param
		// of the block
		
		if(instance->current_block == NULL)
		    // no current block !
    		    remain_time_str[0] = 0;
		else
		    tc_sec2txt
		    (
			(double) (instance->current_block->planned_to.tv_sec - current_time.tv_sec)
			+
			( (double) (instance->current_block->planned_to.tv_usec - current_time.tv_usec) ) / (double)1000000.0 ,
			remain_time_str,
			1
		    );
		
	    }
	    else if ( instance->state == SCHD_PLAYBACK )
	    {
		// in case of playback remains - remains frames param
		tc_frames2txt(instance->playback_remain,remain_time_str);
	    }
	    else
		// in other cases - empty string
		remain_time_str[0] = 0;

	    // DRAW ROW #2 (system state)
	    mvprintw
	    (
		1,0,
		"%10s %10s  %25s  %30s",
		"system: ",
		states_str[instance->state], 
		remain_time_str,
		current_time_str
	    );
	    mvchgat(1, 0, -1, A_NORMAL, 1, NULL);
	    
	    // DRAW ROW #3 (block info)
	    if(instance->current_block == NULL)
	    {
		mvprintw(2,0, "%20s%60s","NO BLOCK PRESENT !!!!","");
		mvprintw(3,0, "%20s%60s","","");
		
	    }
	    else
	    {
		// BLOCK DESCIPTION!
		
		char planned_from[40], planned_to[40], dur[15];
		timeval2tc(instance->current_block->planned_from,planned_from);
		timeval2tc(instance->current_block->planned_to,planned_to);
		
		if((instance->state == SCHD_PLAYBACK) || (instance->state == SCHD_CUE_DONE))
		    tc_frames2txt(instance->playback_length,dur);
		else
		    tc_sec2txt(instance->current_block->duration,dur,1);
	    
    		mvprintw
		(
		    2,0,
		    "%10s <%12s>  [ %24s %24s ] ",
		    "block: ",
		    dur,
		    planned_from,
		    planned_to
		);

    		mvprintw
		(
		    3,0,
		    "#%6d %2s [%12d] %s  %52s",
		    instance->current_block->sys_id,
		    " ",
		    instance->current_block->events_count,
		    instance->current_block->title,
		    " "
		);
		
	    };
	    mvchgat(2, 0, -1, A_NORMAL, 3, NULL);
	    mvchgat(3, 0, -1, A_NORMAL, 3, NULL);

	    // EVENTS DESCRIPTION
    	    mvprintw(4,0,"%10s%70s","events: ","");
	    mvchgat(4, 0, -1, A_NORMAL, 4, NULL);
	    
	    for(i = 0; i< (rows - 6) /*19*/; i++)
	    {
		int f_drawn = 0;
		
		if(instance->current_block)
		    if (instance->current_block->events_count > i)
		    {
			struct schd_event_item* event = &instance->current_block->events[i];

			// do not draw empty spaces
			f_drawn = 1;

			// convert tcs
			char tc[15],dur[15];
			tc_frames2txt(event->tc,tc);
			tc_frames2txt(event->dur,dur);

			// event remain
			char rem[20]; rem[0] = 0;

			if
			(
			    (instance->state == SCHD_CUE_DONE)
			    ||
			    (instance->state == SCHD_PLAYBACK)
			)
			{
			    if( i<instance->playback_count)
				if(instance->playback_remains[i] > 0)
				    tc_frames2txt(instance->playback_remains[i],rem);
			};

			// print			
        		mvprintw
			(
		    	    5 + i,0,
			    "%12s|%10s|%12s|%12s| %s %40s",
			    rem,
			    event->id,
			    tc,
			    dur,
			    event->title,
			    " "
			);
		    };
		
		// draw mpty line
		if(!(f_drawn))
        	    mvprintw(5 + i,0,"%80s"," ");
		
		// setup style
		mvchgat(5+i, 0, -1, A_NORMAL, 4, NULL);

	    };

	    // DRAW BOTTOM ROW (system state aux)
	    char sw_ack[20]; tc_frames2txt(timeval_diff_mili(current_time,instance->f_switch_ack)/40, sw_ack) ;
	    char sw_cmd[20]; tc_frames2txt(timeval_diff_mili(current_time,instance->f_switch_cmd)/40, sw_cmd) ;
	    char cmd_ack[20]; tc_frames2txt(timeval_diff_mili(current_time,instance->f_extcmd_ack)/40, cmd_ack) ;
	    mvprintw
	    (
		rows - 2,0,
		"%7s|%7s|%5s|%7s|%11s|%11s|%11s|%3s|%3s|%3s|%2s",
		"f.ld",
		"f.st",
		"1394",
		"DVOUT",
		"sw-ack",
		"sw-cmd",
		"cmd-ack",
		"sw",
		"cmd",
		"lop",
		"au"
	    );
		
	    mvprintw
	    (
		rows - 1,0,
		"%.7ld|%.7ld|%5s|%7s|%11s|%11s|%11s|%3s|%3s|%3s|%2s",
		instance->cacheman_frames_loaded,
		instance->cacheman_frames_sent,
		(instance->dv_status)?"OK":"FAIL",
		(instance->dv_state)?instance->dv_state:"<none>",
		sw_ack,
		sw_cmd,
		cmd_ack,
		(instance->f_enable_switcher)?"ENA":"DIS",
		(instance->f_enable_extcmd)?"ENA":"DIS",
		(instance->f_loop_block)?"ENA":"DIS",
		(instance->cfg_schd_blocks_aux)?"Y":"N"
	    );
	    mvchgat(rows - 1, 0, -1, A_NORMAL, 2, NULL);
	    mvchgat(rows - 2, 0, -1, A_NORMAL, 2, NULL);


	    pthread_mutex_unlock(&instance->lock_block_read);

	    refresh();

	    // read character
	    long ch;
	    gettimeofday(&current_time, NULL);
	    if(ERR != (ch = getch()))
	    {
		syslog(LOG_NOTICE, "loop_display: key '%c'(0x%.8X) pressed at %ld.%.6ld",(char)ch,ch,current_time.tv_sec,current_time.tv_usec);
		
		// proceed keypress
		switch(ch)
		{
		    case 'a':
		    case 'A':
			instance->cfg_schd_blocks_aux ^= 1;
			break;
		
		    case 'l':
		    case 'L':
			instance->f_loop_block ^= 1;
			break;

		    case 'w':
		    case 'W':
			instance->f_enable_switcher ^= 1;
			break;

		    case 'm':
		    case 'M':
			instance->f_enable_extcmd ^= 1;
			break;
		
		    case 'p':
		    case 'P':
			instance->f_play = 1;
			instance->f_play_occur = current_time;
			break;
			
		    case 's':
		    case 'S':
			instance->f_stop = 1;
			instance->f_stop_occur = current_time;
			break;
			
		    case 'e':
		    case 'E':
			instance->f_stop_emerg = 1;
			instance->f_stop_emerg_occur = current_time;
			break;

		    case 'c':
		    case 'C':
			instance->f_force_cue = 1;
			break;

		};
	    };
	    
	};
	
    };


    
    // setup flag thread active
    instance->f_display = 0;

    // return value
    syslog(LOG_NOTICE, "loop_display: finished");
    pthread_exit(0);
    
    return NULL;
};

int stop_display(struct instance_item* instance)
{
    void* loop_display_exit;


    
    
    if(instance->f_display)
    {
    	syslog(LOG_NOTICE, "stop_display: Waiting for 'display' thread exit");
    	instance->f_display = 0;
    	if(pthread_join(instance->th_display, &loop_display_exit) < 0)
	    syslog(LOG_NOTICE, "stop_display: Something wrong with joining to 'dv_display' thread");
	else
	    syslog(LOG_NOTICE, "stop_display: 'display' exits with '0x%.8X' ", (long)loop_display_exit);
    }
    else
	syslog(LOG_NOTICE, "stop_display: 'display' seems allready finished or not runned");

    // stop ncurses
    endwin();

    // no thread started - exit
    return 1;
};

int start_display(struct instance_item* instance)
{
    setlocale(LC_ALL, "en_US.UTF-8");
//    printf("langinfo='%s'\n",nl_langinfo(CODESET));

    // ncursed init
    initscr();
    nonl();
    noecho();
    cbreak();
    halfdelay(1);
    
    if (has_colors())
    {
    	// init colour usage
    	start_color();
    	
    	// prep colour pairs
    	init_pair(COLOR_BLACK, COLOR_BLACK, COLOR_BLACK);
    	init_pair(COLOR_GREEN, COLOR_GREEN, COLOR_BLACK);
    	init_pair(COLOR_RED, COLOR_RED, COLOR_BLACK);
    	init_pair(COLOR_CYAN, COLOR_CYAN, COLOR_BLACK);
    	init_pair(COLOR_WHITE, COLOR_WHITE, COLOR_BLACK);
    	init_pair(COLOR_MAGENTA, COLOR_MAGENTA, COLOR_BLACK);
    	init_pair(COLOR_BLUE, COLOR_BLUE, COLOR_BLACK);
    	init_pair(COLOR_YELLOW, COLOR_YELLOW, COLOR_BLACK);

	init_pair(1, COLOR_CYAN, COLOR_BLACK);
	init_pair(2, COLOR_WHITE, COLOR_BLUE);
	init_pair(3, COLOR_RED, COLOR_BLACK);
	init_pair(4, COLOR_GREEN, COLOR_BLACK);
    };
    
    printw("[%2d]%30s%20s%26s",instance->cfg_instance_no, " Sheduled Video Replacer: ", VERSION, " ");
    mvchgat(0, 0, -1, A_NORMAL, 2, NULL);


    refresh();

    if(pthread_create(&instance->th_display, &instance->th_display_attr, loop_display, instance) < 0)
    {
	syslog(LOG_ERR, "start_display: Unable to start 'display' thread!");
	return 0;
    }
    else
	return 1;
};
