/*

Changelog:
    2005-11-11:
	*Threads priority adjustment.

    2005-11-10:
	*unpluging avc device do not cause increasing dropped frames count (may be because
	we do not use DV1394_IOC_WAIT_FRAMES). Watchdog timeof_DVOUT_STATE_COPY_DATA
	added to detect dropped frames manualy.
	
    2005-11-07:
	*decreased length of state text description up to 7 characters
    
    2005-11-05 17:00
	*Solved problem with DV1394_IOC_WAIT_FRAMES. It's seems something wrong with it
	it too late to return. System from now can work in 'mmap()'ed mode on in 'write()'
	mode depends on 'DVOUT_USE_MMAP'.

    2005-11-05 04:00
	*Almost done!. Fake module rewriten to support DV1394 IOC scheme using FSM,
	*Added audio muting for still frames.
	
*/
	    

#include <unistd.h>
#include <stdio.h>
#include <syslog.h>
#include <time.h>
#include <string.h>

#include <asm/ioctls.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>

#include <drivers/ieee1394/dv1394.h>
#include <drivers/ieee1394/ieee1394-ioctl.h>

#include "schd_replacer.h"
#include "cacheman.h"
#include "statectl.h"

//#define VERBOSE_STATE_LOG
//#define VERBOSE_STATE_LOG2



#define DVOUT_CLEAR_FRAMES_TIMEOUT 80

#define DVOUT_USE_MMAP
//#define DVOUT_USE_WAIT_FRAMES		/* DO NOT UNCOMMENT !!!!! */
#define DVOUT_USE_MMAP_FO 0 		/* strange situation - we should fill frame after one ... */


#define DVOUT_MAX_RING_BUFFERS 10
//#define DVOUT_RING_BUFFERS 2
#define DVOUT_RING_BUFFERS (instance->cfg_dv_ringbuffers)
#define DVOUT_ERRSLEEP_NS 0 /* (25*40*1000) */
#define DVOUT_ERRSLEEP_S 1 

static int dv_dev = 0;

static void* mapped_buffers[DVOUT_MAX_RING_BUFFERS];
static void* mapped_buffer;

static struct dv1394_init dv_init_data;


// lib dv operations
#include "libdv/dv.h"
#include "libdv/enc_audio_input.h"
#include "libdv/enc_output.h"
static dv_enc_audio_info_t MUTE;


enum DVOUT_STATES
{
    DVOUT_STATE_OPEN_DEV = 0,
    DVOUT_STATE_INIT_DEV,
    DVOUT_STATE_MAP_BUFFERS,
    DVOUT_STATE_WAIT_FRAMES,
    DVOUT_STATE_GET_STATUS,
    DVOUT_STATE_COPY_DATA,
    DVOUT_STATE_SUBMIT_FRAMES,
    DVOUT_STATE_UNMAP_BUFFERS,
    DVOUT_STATE_ERRSLEEP,
    DVOUT_STATE_CLOSE_DEV
};

// initial state
DVOUT_STATES state = DVOUT_STATE_OPEN_DEV;

    
static char* dvout_states_txt[] = 
{
    "OPENDEV",
    "INITDEV",
    "MAPBUFS",
    "WAITFRM",
    "GETSTAT",
    "CPYDATA",
    "SUBMITF",
    "UMAPBUF",
    "ERRSLP",
    "CLOSEDEV"
};


// main dv output loop
static void* loop_dvout(void* p)
{
    // cast instance ctrl struct
    struct instance_item* instance = (struct instance_item*)p;

    // proc vars
    long error, i, l;
    void* frame_data_ptr = NULL , ;
    struct dv1394_status dv_status;
    struct timeval current_time;

    // mute control
    long index_prev = 0, index_current = 0, index_muted = (-1);
    memset(&MUTE,0,sizeof(dv_enc_audio_info_t));
    MUTE.channels	= 2;
    MUTE.frequency	= 48000;
    MUTE.bitspersample	= 16;
    MUTE.bytealignment	= 2;
    MUTE.bytespersecond	= MUTE.frequency*MUTE.channels*MUTE.bytealignment;
    MUTE.bytesperframe	= MUTE.bytespersecond/25;
    memset(MUTE.data,0,sizeof(MUTE.data));
    

    // setup init struct
    dv_init_data.api_version 	= DV1394_API_VERSION;	/* DV1394_API_VERSION */
    
    dv_init_data.channel	= 0x63;			/* broadcast channel */ 
							/* isochronous transmission channel to use */

    dv_init_data.n_frames	= DVOUT_RING_BUFFERS;	/* ringbuffer length */
							/* number of frames in the ringbuffer. Must be at least 2
							    and at most DV1394_MAX_FRAMES. */
			    
    dv_init_data.format		= DV1394_PAL;		/* send PAL video */
							/* send/receive PAL or NTSC video format */
    
    /* the following are used only for transmission */
    dv_init_data.cip_n		= 0; 			/* default empty packet rate */
    dv_init_data.cip_d		= 0;
    dv_init_data.syt_offset	= 0;

    /* reset buffer ptr */
    mapped_buffer = MAP_FAILED;
    

    // setup flag thread active
    instance->f_dvout = 1;
    syslog(LOG_NOTICE, "loop_dvout: started");

    // initial state
    state = DVOUT_STATE_OPEN_DEV;

    // sleep before close
    struct timespec wf_sleep_time;
    wf_sleep_time.tv_sec = 0; wf_sleep_time.tv_nsec = 1000;	// one milisecond to sleep !
    struct timespec wf_slept_time;

    // error case timeout
    struct timespec errsleep_time;
    errsleep_time.tv_sec = DVOUT_ERRSLEEP_S; errsleep_time.tv_nsec =  DVOUT_ERRSLEEP_NS;
    struct timespec errslept_time;

    // update text state info
    instance->dv_state = dvout_states_txt[state];
    instance->dv_status = 0;

    // last frame copied
    struct timeval timeof_DVOUT_STATE_COPY_DATA;
    gettimeofday(&timeof_DVOUT_STATE_COPY_DATA, NULL);

    // program loop
    while	/* we exit when exit flag rized and FSM in 
                   DVOUT_STATE_OPEN_DEV state that mean
                   that no open devices here and mapped
                   buffers */
    (!(
    	(state == DVOUT_STATE_OPEN_DEV)
	&&
	(
	    (!(instance->f_dvout))
	    ||
            (instance->f_exit)
        )
    ))    
    {            
	// fix time
	gettimeofday(&current_time, NULL);

	// update text state info
	instance->dv_state = dvout_states_txt[state];

#ifdef VERBOSE_STATE_LOG
	syslog(LOG_NOTICE, "loop_dvout: [%s] VERBOSE_STATE_LOG  at %ld.%.6ld",
	    dvout_states_txt[state],
	    current_time.tv_sec,current_time.tv_usec);
#endif	

	// FSM begins
    	switch(state)
    	{
            case DVOUT_STATE_OPEN_DEV:
		gettimeofday(&timeof_DVOUT_STATE_COPY_DATA, NULL);

            	// try to open device
            	if( (dv_dev = open(instance->cfg_dv_dev_name, O_RDWR))<0 )
		{
		    syslog(LOG_NOTICE, "loop_dvout: [%s] failed to open device '%s'  at %ld.%.6ld",
			dvout_states_txt[state],
			instance->cfg_dv_dev_name,
			current_time.tv_sec,current_time.tv_usec);
		    state = DVOUT_STATE_ERRSLEEP;
		}
		else
		{
		    syslog(LOG_NOTICE, "loop_dvout: [%s] success open device '%s'  at %ld.%.6ld",
			dvout_states_txt[state],
			instance->cfg_dv_dev_name,
			current_time.tv_sec,current_time.tv_usec);
		    state = DVOUT_STATE_INIT_DEV;
		};
		break;


            	
            case DVOUT_STATE_INIT_DEV:
		// init device
		if((error = ioctl(dv_dev, DV1394_IOC_INIT, &dv_init_data)) < 0 )
		{
		    syslog(LOG_NOTICE, "loop_dvout: [%s] failed to DV1394_IOC_INIT result: '%d'  at %ld.%.6ld",
			dvout_states_txt[state],
			error,
			current_time.tv_sec,current_time.tv_usec);
		    state = DVOUT_STATE_ERRSLEEP;
		}
		else
		{
		    syslog(LOG_NOTICE, "loop_dvout: [%s] success DV1394_IOC_INIT at %ld.%.6ld",
			dvout_states_txt[state],
			current_time.tv_sec,current_time.tv_usec);
		    state = DVOUT_STATE_MAP_BUFFERS;
		};
		break;


		
            case DVOUT_STATE_MAP_BUFFERS:
#ifdef DVOUT_USE_MMAP
		// try to map buffer
		if( (mapped_buffer = mmap(NULL, DV1394_PAL_FRAME_SIZE*DVOUT_RING_BUFFERS , PROT_WRITE , MAP_SHARED, dv_dev, 0)  )  == MAP_FAILED)
		{
		    syslog(LOG_NOTICE, "loop_dvout: [%s] failed to map buffer at %ld.%.6ld",
			dvout_states_txt[state],
			current_time.tv_sec,current_time.tv_usec);
		    state = DVOUT_STATE_UNMAP_BUFFERS;
		}
		else
		{
		    // recalc buffers offset
		    for(i=0;i<DVOUT_RING_BUFFERS;i++)
			mapped_buffers[i] = (void*)(((unsigned long)mapped_buffer)+i*DV1394_PAL_FRAME_SIZE);
		    state = DVOUT_STATE_WAIT_FRAMES;
		};
		break;    
#else
		state = DVOUT_STATE_WAIT_FRAMES;
		break;
#endif


	    
            case DVOUT_STATE_WAIT_FRAMES:
#ifdef DVOUT_USE_WAIT_FRAMES
                // wait for frames using IOC DV1394_IOC_WAIT_FRAMES
                if((error = ioctl(dv_dev, DV1394_IOC_WAIT_FRAMES, 1)) < 0 )
                {
                    syslog(LOG_NOTICE, "loop_dvout: [%s] failed DV1394_IOC_WAIT_FRAMES result: '%d'  at %ld.%.6ld",
                	dvout_states_txt[state],
                	error,
                	current_time.tv_sec,current_time.tv_usec);
		    state = DVOUT_STATE_UNMAP_BUFFERS;
		}
		else
		    state = DVOUT_STATE_GET_STATUS;
#else
		state = DVOUT_STATE_GET_STATUS;
#endif	// DVOUT_USE_WAIT_FRAMES

		break;


	    
            case DVOUT_STATE_GET_STATUS:
		// get status
		if((error = ioctl(dv_dev, DV1394_IOC_GET_STATUS, &dv_status)) < 0 )
		{
		    syslog(LOG_NOTICE, "loop_dvout: [%s] failed DV1394_IOC_GET_STATUS result: '%d'  at %ld.%.6ld",
			dvout_states_txt[state],
			error,
			current_time.tv_sec,current_time.tv_usec);
		    state = DVOUT_STATE_UNMAP_BUFFERS;
		}
		else
		{

#ifdef VERBOSE_STATE_LOG1
		syslog(LOG_NOTICE, "loop_dvout: dv_status.dropped_frames=%d,dv_status.n_clear_frames=%d  at %ld.%.6ld",
		    dv_status.dropped_frames,
		    dv_status.n_clear_frames,
		    current_time.tv_sec,current_time.tv_usec);
#endif	

		    // check for dropped frames
		    if
		    (
		    	(dv_status.dropped_frames>0)
		    	||
		    	(timeval_diff_mili(current_time,timeof_DVOUT_STATE_COPY_DATA) > DVOUT_CLEAR_FRAMES_TIMEOUT)
		    )
		    {
			syslog(LOG_NOTICE, "loop_dvout: [%s] dropped_frames='%d'  at %ld.%.6ld",
			    dvout_states_txt[state],
			    dv_status.dropped_frames,
			    current_time.tv_sec,current_time.tv_usec);
			state = DVOUT_STATE_UNMAP_BUFFERS;
		    }
		    else
		    {
#ifdef DVOUT_USE_MMAP

#ifndef DVOUT_USE_WAIT_FRAMES
			// check if number if clear frames
			if(dv_status.n_clear_frames > 0)
			    state = DVOUT_STATE_COPY_DATA;
			else
            		    nanosleep(&wf_sleep_time,&wf_slept_time); 
#else /* DVOUT_USE_WAIT_FRAMES defined  */
			state = DVOUT_STATE_COPY_DATA;
#endif /* DVOUT_USE_WAIT_FRAMES*/

#else /* DVOUT_USE_MMAP not defined */
		        state = DVOUT_STATE_COPY_DATA;
#endif /* DVOUT_USE_MMAP */
		    };
		};
		break;
	    
            case DVOUT_STATE_COPY_DATA:
            
            	// fix time of this state
            	gettimeofday(&timeof_DVOUT_STATE_COPY_DATA, NULL);
            
		instance->dv_status = 1;
		// check for exit condition
		if
		(
		    (!(instance->f_dvout))
		    ||
        	    (instance->f_exit)
		)
		{
		    state = DVOUT_STATE_UNMAP_BUFFERS;
		    syslog(LOG_NOTICE, "loop_dvout: [%s] exiting...  at %ld.%.6ld",
			dvout_states_txt[state],
			current_time.tv_sec,current_time.tv_usec);
		}
		else
		{	
		    // retrive frame data ptr
		    frame_data_ptr = get_active_frame(instance, &index_current);

		    // check for NULL
		    if
		    (
			(frame_data_ptr)
//			&&
//			(instance->state != SCHD_CUE_INIT)
		    )
		    {
			// insert timecode ?
			dv_encode_timecode((unsigned char*)frame_data_ptr, 1, instance->cacheman_frames_loaded);
			
#ifdef VERBOSE_STATE_LOG2
            		gettimeofday(&current_time, NULL);
			syslog
			(
			    LOG_NOTICE, 
			    "loop_dvout: index_current=%d, index_prev=%d, index_muted=%d at %ld.%.6ld",
			    index_current,
			    index_prev,
			    index_muted,
			    current_time.tv_sec,current_time.tv_usec
			);
#endif
		    
			// check for audio mute	
			if(index_current == index_prev) 
			{
			    // previous frame had the same index
			    
			    // has we mute it?
			    if(index_current != index_muted)
			    {
				// set index of muted frame
				index_muted = index_current;
				
				// mute video in cache buffer
				_dv_raw_insert_audio((unsigned char*)frame_data_ptr, &MUTE, 1);
#ifdef VERBOSE_STATE_LOG2
            			gettimeofday(&current_time, NULL);
				syslog(LOG_NOTICE, "loop_dvout: muted 0x%.8x at %ld.%.6ld",
				    (long)frame_data_ptr,
				    current_time.tv_sec,current_time.tv_usec);
#endif
			    };
			}
			else
			{
			    index_muted = (-1);
			    index_prev = index_current;
			};
		    
		    
			// copy data
#ifdef DVOUT_USE_MMAP
			int j = (dv_status.first_clear_frame + DVOUT_USE_MMAP_FO) % DVOUT_RING_BUFFERS;
			memmove(mapped_buffers[j] ,frame_data_ptr, DV1394_PAL_FRAME_SIZE);
#ifdef VERBOSE_STATE_LOG
            		gettimeofday(&current_time, NULL);
			syslog(LOG_NOTICE, "loop_dvout: memmove 0x%.8x->%d at %ld.%.6ld",
			    (long)frame_data_ptr,
			    j,
			    current_time.tv_sec,current_time.tv_usec);
#endif

#else
			write(dv_dev, frame_data_ptr, DV1394_PAL_FRAME_SIZE);
#endif			
		    };
//		    else
//		    {
//			syslog(LOG_NOTICE, "loop_dvout: [%s] NULL ptr   at %ld.%.6ld",
//			    dvout_states_txt[state],
//			    current_time.tv_sec,current_time.tv_usec);
//		    };
		    
		    // release frame
		    mark_active_frame_clear(instance);


		    state = DVOUT_STATE_SUBMIT_FRAMES;
		};
		break;


	    
            case DVOUT_STATE_SUBMIT_FRAMES:
#ifdef DVOUT_USE_MMAP
		// submit frames (actualy one)
		if((error = ioctl(dv_dev, DV1394_IOC_SUBMIT_FRAMES, 1)) < 0 )
		{
		    syslog(LOG_NOTICE, "loop_dvout: [%s] failed DV1394_IOC_SUBMIT_FRAMES result: '%d'  at %ld.%.6ld",
			dvout_states_txt[state],
			error,
			current_time.tv_sec,current_time.tv_usec);
		    state = DVOUT_STATE_UNMAP_BUFFERS;
		}
		else
#endif
		    state = DVOUT_STATE_GET_STATUS;
		break;


	    
            case DVOUT_STATE_UNMAP_BUFFERS:
    		syslog(LOG_NOTICE, "loop_dvout: [%s] exiting...  at %ld.%.6ld",
		    dvout_states_txt[state],
		    current_time.tv_sec,current_time.tv_usec);
		instance->dv_status = 0;
#ifdef DVOUT_USE_MMAP
		if(mapped_buffer != MAP_FAILED)
		{
		    munmap(mapped_buffer, DV1394_PAL_FRAME_SIZE * DVOUT_RING_BUFFERS);
		    mapped_buffer = MAP_FAILED;
		};
#endif
		state = DVOUT_STATE_ERRSLEEP;
		break;



            case DVOUT_STATE_ERRSLEEP:
            	nanosleep(&errsleep_time,&errslept_time); 
            	state = DVOUT_STATE_CLOSE_DEV; 
    		syslog(LOG_NOTICE, "loop_dvout: [%s] exiting...  at %ld.%.6ld",
		    dvout_states_txt[state],
		    current_time.tv_sec,current_time.tv_usec);
            	break;
                
            case DVOUT_STATE_CLOSE_DEV:
            	// close device if it opened
                if(dv_dev > 0) close(dv_dev);
                state = DVOUT_STATE_OPEN_DEV; 
                break;
    	};
    };

    // setup flag thread active
    instance->f_dvout = 0;

    // return value
    syslog(LOG_NOTICE, "loop_dvout: finished");
    pthread_exit(0);
    
    return NULL;
};

int stop_dvout(struct instance_item* instance)
{
    void* loop_dvout_exit;
    
    if(instance->f_dvout)
    {
    	syslog(LOG_NOTICE, "stop_dvout: Waiting for 'dv_out' thread exit");
    	instance->f_dvout = 0;
    	if(pthread_join(instance->th_dvout, &loop_dvout_exit) < 0)
	    syslog(LOG_NOTICE, "stop_dvout: Something wrong with joining to 'dv_out' thread");
	else
	    syslog(LOG_NOTICE, "stop_dvout: 'dv_out' exits with '0x%.8X' ", (long)loop_dvout_exit);
    }
    else
	syslog(LOG_NOTICE, "stop_dvout: 'dv_out' seems allready finished or not runned");
    	// no thread started - exit
        return 1;
};

int start_dvout(struct instance_item* instance)
{
    if(pthread_create(&instance->th_dvout, &instance->th_dvout_attr, loop_dvout, instance) < 0)
    {
	syslog(LOG_ERR, "start_dvout: Unable to start 'dv_out' thread!");
	return 0;
    }
    else
	return 1;
};
