/*

Changelog:
    2005-11-17:
	*Sleeping before rising flag added
	
    2005-11-11:
	*Threads priority adjustment.

    2005-11-07:
	*Fixed 'FD_SET/select' block, select() returns number of
	DESCRIPTORS! Added section which detect number of recieved bytes.
	*'O_NDELAY' option added to port open option bits.
	
    2005-11-06:
	*Enable 'remote/local' control using f_enable_extcmd

    2005-11-04 04:00:00
	*Seems model is working !!! Let's test and fix!
	
*/


#include <unistd.h>
#include <stdio.h>
#include <syslog.h>
#include <time.h>
#include <fcntl.h>
#include <termios.h>
#include <strings.h>

#include "schd_replacer.h"

static struct termios oldtio,newtio;
static int serial_handler;


/*

Команды управления имеют фиксированную длину 4 байта. Фомат команд:
     02    01    01    FF
    ----  ----  ----  ----
    STX   CMD1  CMD2  ETX
где
    STX - стартовый заголовок
    CMD1,CMD2 - 2х байтный код команды
    ETX - окончание команды

В ответ ваша система посалает 2-х байтный ответ:
    04    00   - ACK
    05    ХХ   - NACK (где ХХ код ошибки) 

Список возможных команд (наличие признака):
    CMD1  CMD2
    ----  ----
    01    01   - вход НА "замещение"
    02    01   - возврат на главный канал

На будущее можно будет раширять диапазон команд для обработки вашем 
програмным обеспечением задачи различной бизнес логики (например 
"обязательный|необязательный" возрат на основной канал и т.п.) 


    0x FF 01 01 02
        ^  ^  ^  ^
        |  |  |  +- STX
        |  |  +---- CMD1 |
	|  +------- CMD2 | - command code
	+---------- ETX
	
*/

#define EXTCMD_CMD_CTL_MASK	0xFF0000FF
#define EXTCMD_CMD_CTL		0xFF000002
#define EXTCMD_CMD_DATA_MASK	0x00FFFF00
#define EXTCMD_CMD_PLAY		0x00010100
#define EXTCMD_CMD_STOP		0x00010200
#define EXTCMD_TIMEOUT		1000*5	/* microseconds */


// extcmd loop
static void* loop_extcmd(void* p)
{
    long data, r, reply;
    fd_set fds;
    struct timeval tv;

    // cast instance ctrl struct
    struct instance_item* instance = (struct instance_item*)p;

    // setup flag thread active
    instance->f_extcmd = 1;
    syslog(LOG_NOTICE, "loop_extcmd: started");

    //
    struct timeval current_time;
    struct timespec sleep_time;
    sleep_time.tv_sec = 0; sleep_time.tv_nsec =  EXTCMD_TIMEOUT;
    struct timespec slept_time;

    // cmd sending logging
    gettimeofday(&instance->f_extcmd_cmd,NULL);
    gettimeofday(&instance->f_extcmd_ack,NULL);

    
    while((instance->f_extcmd)&&(!(instance->f_exit)))
    {
        // get current time
	gettimeofday(&current_time,NULL);
                
	// check if device open
	if (serial_handler > 0)
	{
	    // async read
            FD_ZERO (&fds);
            FD_SET (serial_handler, &fds);
            tv.tv_sec = 0; 
            tv.tv_usec = EXTCMD_TIMEOUT; 
            long r = select (serial_handler + 1, &fds, NULL, NULL, &tv);
	    gettimeofday(&current_time,NULL);
	    
	    // check what we got
	    if(r>0)
	    {
		// read data
		data = 0;
		r = read(serial_handler, &data, 4);
		gettimeofday(&current_time,NULL);
	    
		// we got some bytes
                if(r == 4)
                {
		    // we got a full length !
		    
		    // check command
		    if( (data & EXTCMD_CMD_CTL_MASK) == EXTCMD_CMD_CTL )
		    {
			gettimeofday(&instance->f_extcmd_cmd,NULL);

			// detect reply and send it
			switch( (data & EXTCMD_CMD_DATA_MASK) )
			{
			    case EXTCMD_CMD_PLAY: reply = 0x04;	break;
			    case EXTCMD_CMD_STOP: reply = 0x04; break;
			    default: reply = 0x0105; break;
			};
			
			// send reply
			write(serial_handler, &reply, 2);
			instance->f_extcmd_ack = instance->f_extcmd_cmd;
			
			// detect command and exec and delay
			struct timespec delay_time = {0,0};
			if(instance->f_enable_extcmd)
			switch( (data & EXTCMD_CMD_DATA_MASK) )
			{
			    case EXTCMD_CMD_PLAY:
				// delay
				delay_time.tv_sec = instance->cfg_extcmd_delay_play/1000 ; 
				delay_time.tv_nsec = (instance->cfg_extcmd_delay_play%1000)*1000000;
				nanosleep(&delay_time, &slept_time);
				// exec
				instance->f_play = 1;
				gettimeofday(&instance->f_play_occur,NULL);
				break;
				
			    case EXTCMD_CMD_STOP:
				// delay
				delay_time.tv_sec = instance->cfg_extcmd_delay_stop/1000 ; 
				delay_time.tv_nsec = (instance->cfg_extcmd_delay_stop%1000)*1000000;
				nanosleep(&delay_time, &slept_time);
				// exec
				instance->f_stop = 1;
				gettimeofday(&instance->f_stop_occur,NULL);
				break;
			};
		
	

			// log it
			syslog(LOG_NOTICE, "loop_extcmd: cmd proceeded [%.8X->%.8X] at %ld.%.6ld, (delay %ld.%.9ld)", 
			    data,reply,current_time.tv_sec,current_time.tv_usec, delay_time.tv_sec, delay_time.tv_nsec);
		    }
		    else
			syslog(LOG_NOTICE, "loop_extcmd: incorrect command struct (%.8X bytes) at %ld.%.6ld", data,current_time.tv_sec,current_time.tv_usec);
		}
		else
		{
		    // we got partial length
		    syslog(LOG_NOTICE, "loop_extcmd: partial/overflow (%d bytes) content recieved at %ld.%.6ld",r,current_time.tv_sec,current_time.tv_usec);
		}
	    };
	    
    	    // flush data in INPUT buffer
//    	    tcflush(serial_handler, TCIFLUSH);
	}
	else
    	    // sleep a bit
    	    nanosleep(&sleep_time, &slept_time);
    };

    // setup flag thread active
    instance->f_extcmd = 0;

    // return value
    syslog(LOG_NOTICE, "loop_extcmd: finished");
    pthread_exit(0);
    
    return NULL;
};

int stop_extcmd(struct instance_item* instance)
{
    void* loop_extcmd_exit;
    
    if (serial_handler >0)
    {
        // restore serial device params
        tcflush(serial_handler, TCIFLUSH);
        tcsetattr(serial_handler,TCSANOW,&newtio);
    };
    
    
    if(instance->f_extcmd)
    {
    	syslog(LOG_NOTICE, "stop_extcmd: Waiting for 'external cmd detector' thread exit");
    	instance->f_switcher = 0;
    	if(pthread_join(instance->th_extcmd, &loop_extcmd_exit) < 0)
	    syslog(LOG_NOTICE, "stop_extcmd: Something wrong with joining to 'external cmd detector' thread");
	else
	    syslog(LOG_NOTICE, "stop_extcmd: 'external cmd detector' exits with '0x%.8X' ", (long)loop_extcmd_exit);
    }
    else
	syslog(LOG_NOTICE, "stop_extcmd: 'external cmd controller' seems allready finished or not runned");

    // no thread started - exit
    return 1;
};

int start_extcmd(struct instance_item* instance)
{
    // open serial device
    serial_handler = 0;
    
    if( (serial_handler = open(instance->cfg_extcmd_device_name, O_RDWR | O_NOCTTY | O_NDELAY ) ) < 0 )
    {
    	syslog(LOG_ERR, "start_extcmd: Unable to open '%s'!", instance->cfg_extcmd_device_name);
    }
    else
    {
    	syslog(LOG_NOTICE, "start_extcmd: Device '%s' opened successful!",instance->cfg_extcmd_device_name);
    	
    	// configure serial device
        tcgetattr(serial_handler,&oldtio);        /* save current serial port settings */
        bzero(&newtio, sizeof(newtio));     /* clear struct for new port settings */
        newtio.c_iflag = IGNPAR;            /* ignore bytes with parity errors */
        newtio.c_oflag = 0;                 /* Raw output */
        newtio.c_lflag = 0;//ICANON;        /* enable canonical input */
        newtio.c_cc[VTIME]    = 0;          /* inter-character timer unused */
        newtio.c_cc[VMIN]     = 1;          /* blocking read until 5 chars received */
        newtio.c_cflag =
	    B38400 |
            PARENB |
            PARODD |
            CS8 |
            CREAD |
            CLOCAL;
        tcflush(serial_handler, TCIOFLUSH);
        tcsetattr(serial_handler,TCSANOW,&newtio);
    };
                                                                                            


    if(pthread_create(&instance->th_extcmd, &instance->th_extcmd_attr, loop_extcmd, instance) < 0)
    {
	syslog(LOG_ERR, "start_extcmd: Unable to start 'external cmd controller' thread!");
	return 0;
    }
    else
	return 1;
};
