#include "mytimecode.h"


#define FPS 25.0

static int is_digit(char c)
{
	return ((c>='0')&&(c<='9'))?1:0;
};

static unsigned char tc_alphabet[16] = 
{
	'0','1','2','3','4','5','6','7',
	'8','9','?','?','?','?','?','?'
};

unsigned long tc_txt2bcd(void* tc_txt)
{
	return tc_frames2bcd(tc_txt2frames(tc_txt));
};

unsigned long tc_txt2frames(void* tc_txt)
{
	char* txt = (char*)tc_txt;

	unsigned char part_found[4] = {0,0,0,0};
	unsigned long part_tc[4] = {0,0,0,0};
	unsigned long part_no = 0;
	while((*txt)&&(part_no < 4))
	{
		if(is_digit(*txt))
		{
			part_found[part_no] = 1;
			part_tc[part_no] *= 10;
			part_tc[part_no] += (unsigned long)(*txt - '0');
			txt++;
			if(!is_digit(*txt))
				part_no++;
		}
		else
		{
			txt++;
		};
	};

	return (part_tc[0]*3600 + part_tc[1]*60 + part_tc[2])*25 + part_tc[3];
};

static unsigned long tc_bcd2frames_mults[8] = 
{
	10 * 3600 * (int)FPS,
	 1 * 3600 * (int)FPS,
	10 *   60 * (int)FPS,
	 1 *   60 * (int)FPS,
	10 *    1 * (int)FPS,
	 1 *    1 * (int)FPS,
	10 *    1 * 1,
	 1 *    1 * 1
};


unsigned long tc_bcd2frames(unsigned long tc_bcd)
{
	unsigned long temp_tc = 0;
	for(int i=0;i<8;i++,tc_bcd>>=4)
		temp_tc+=tc_bcd2frames_mults[7-i]*(tc_bcd&0xf);
	return temp_tc;
};

unsigned long tc_frames2bcd(unsigned long tc_frames)
{
	unsigned long divs[4] = {(int)FPS,60,60,99};
	unsigned long temp_tc = 0;
	unsigned long tc_part;
	for(int i=0;i<4;i++)
	{
		tc_part = tc_frames % divs[i];
		tc_frames /= divs[i];
		temp_tc |= ((tc_part/10)<<4 | (tc_part%10))<<(i*8);
	};
	return temp_tc;
};


char* tc_bcd2txt(unsigned long tc_bcd, void* tc_txt_buf)
{
	// case
	unsigned char* txt = (unsigned char*)tc_txt_buf;

	// finish
	txt[11] = 0;

	// frames
	txt[10] = tc_alphabet[ tc_bcd & 0xF ];tc_bcd>>=4;
	txt[9] = tc_alphabet[ tc_bcd & 0xF ];tc_bcd>>=4;
	txt[8] = ':';

	// seconds
	txt[7] = tc_alphabet[ tc_bcd & 0xF ];tc_bcd>>=4;
	txt[6] = tc_alphabet[ tc_bcd & 0xF ];tc_bcd>>=4;
	txt[5] = ':';

	// minutes
	txt[4] = tc_alphabet[ tc_bcd & 0xF ];tc_bcd>>=4;
	txt[3] = tc_alphabet[ tc_bcd & 0xF ];tc_bcd>>=4;
	txt[2] = ':';

	// minutes
	txt[1] = tc_alphabet[ tc_bcd & 0xF ];tc_bcd>>=4;
	txt[0] = tc_alphabet[ tc_bcd & 0xF ];tc_bcd>>=4;

	return (char*)tc_txt_buf;
};


char* tc_frames2txt(unsigned long tc_frames, void* tc_txt_buf)
{
	unsigned long tc_bcd = tc_frames2bcd(tc_frames);
	return tc_bcd2txt(tc_bcd,tc_txt_buf);
};

char* tc_sec2txt(double tc_sec, void* tc_txt_buf, int wrap_day)
{
	long frames = (long)( tc_sec  * (double)25.0 );
	if(wrap_day)
		frames = frames % (24*3600*25);
	return tc_frames2txt(frames, tc_txt_buf);
};
