/*

Changelog:
    2005-11-11:
	*Threads priority adjastment.

    2005-11-09:
	*sys_id parameter added to block struct and modified load_block proc
	*black_list functionality to avoid cueing played items and solving 
	problem with overlapped region 
	
    2005-11-07:
	*Schedule reload forbidden in SCHD_POSTROLL mode
	
    2005-11-04:
	*More smart scheme of schedule reload

    2005-11-03:
	*First released version

*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>

#include "schd.h"
#include "schd_replacer.h"

static long* blacklist_blocks_sys_id;
static long blacklist_blocks_count;

static int schd_event_load(char* buf, struct schd_event_item* event, long line_no)
{
    long tc_h, tc_m, tc_s, tc_f, dur_h, dur_m, dur_s, dur_f;
	    
    // try to recongnize
    if
    (
	10 == sscanf
	(
	    buf,
	    "%s %d:%d:%d:%d %d:%d:%d:%d %[^\n]",
	    event->id,
	    &tc_h,&tc_m,&tc_s,&tc_f,
	    &dur_h,&dur_m,&dur_s,&dur_f,
	    event->title
	)
    )
    {
	// calc tc in frames
	event->tc = (tc_h*3600 + tc_m*60 + tc_s)*25 + tc_f;
	event->dur = (dur_h*3600 + dur_m*60 + dur_s)*25 + dur_f;
	return 1;
    }
    else
    {
	syslog(LOG_NOTICE, "schd_event_load: line [%d] '%s' not recongnized as event item!!!!!\n", line_no, buf);
	return 0;
    };
};



static int schd_block_load(char* buf, struct schd_block_item* block, long line_no)
{
    struct tm p_t;
    struct tm p_f;
    long f_f,f_t;

    // clear block
    memset(block,0,sizeof(struct schd_block_item));

    // clear time structs    
    memset(&p_t,0,sizeof(struct tm));
    memset(&p_f,0,sizeof(struct tm));

    // try to recongnize
    if
    (
	16 == sscanf
	(
	    buf,
	    "%d-%d-%d %d:%d:%d:%d   %d-%d-%d %d:%d:%d:%d  [%d]  %[^\n]",

	    &p_f.tm_year, &p_f.tm_mon, &p_f.tm_mday,
		&p_f.tm_hour, &p_f.tm_min, &p_f.tm_sec,
		    &f_f,
		
	    &p_t.tm_year, &p_t.tm_mon, &p_t.tm_mday,
	    	&p_t.tm_hour, &p_t.tm_min, &p_t.tm_sec,
		    &f_t,

	    &block->sys_id,

	    block->title
	)
    )
    {
	// normalize
	p_t.tm_year -= 1900; p_t.tm_mon -= 1;
	p_f.tm_year -= 1900; p_f.tm_mon -= 1;
	
	// assign values
	block->planned_from.tv_sec = mktime(&p_f); block->planned_from.tv_usec = 1000 * 40 * f_f;
	block->planned_to.tv_sec = mktime(&p_t); block->planned_to.tv_usec = 1000 * 40 * f_t;
    
	return 1;
    }
    else
    {
	syslog(LOG_NOTICE, "schd_block_load: line [%d] '%s' not recongnized as block item!!!!!\n", line_no, buf);
	return 0;
    };
};

static int schd_build(struct schd_block_item* &blocks, FILE* f)
{
    long i,j,l;
    
    // string buffer
    char buf[1024];
    
    // setup counter
    int count = 0;
    
    // allocate space for array
    blocks = (struct schd_block_item*)malloc(0);
    
    l = 0;
    while(!(feof(f)))
    {
	// read string
	fgets(buf, sizeof(buf) - 1, f);
	    
	// detect is it item block begins or schedule
	if(buf[0] == '*')
	{
	    struct schd_block_item block;
	    
	    // try to load
	    if(schd_block_load(buf+1,&block,l))
	    {
		// block loaded !!
		
		// increment counter
		count++;
		
		// realloc space for block
		blocks = (struct schd_block_item*)realloc(blocks, count*sizeof(struct schd_block_item));
		
		// save block
		blocks[count - 1] = block;
		
		// alloc space for events
		blocks[count - 1].events = (struct schd_event_item*) malloc(0);
	    };
	}
	else if(buf[0] == '+')
	{
	    struct schd_event_item event;
	    
	    // try to load event
	    if(schd_event_load(buf+1,&event,l))
	    {
		// event loaded !!
		
		// increment counter
		blocks[count - 1].events_count++;
		
		// realloc space for event
		blocks[count - 1].events = (struct schd_event_item*)realloc(blocks[count - 1].events, blocks[count - 1].events_count*sizeof(struct schd_event_item));
		
		// save block
		blocks[count - 1].events[ blocks[count - 1].events_count - 1]  = event;
	    };
	        
	};
	
	// increment line no
	l++;
    };

    // calc durations
    for(i=0;i<count;i++)
    {
	// sum items
	for(l=0,j=0;j<blocks[i].events_count;j++) l+=blocks[i].events[j].dur;
	blocks[i].duration = (double)l / (double)25.0 ;
    };
    
    // count
    return count;
};


static int schd_free(struct schd_block_item* &blocks, long count)
{
    int i, j;
    
    if(blocks)
    {
	for(int i=0;i<count;i++)
	    free(blocks[i].events);
	free(blocks);
	
	blocks = NULL;
    };
};


static void schd_print(struct schd_block_item* blocks, long blocks_count)
{
    return;
    
    printf("Blocks count: %d\n", blocks_count);
    
    // check for count and print it
    if( (blocks_count) && (blocks) )
    {
        for(int i=0;i<blocks_count;i++)
        {
	    struct tm temp_tm;
	    char buf1[128], buf2[128];

	    // build text representation of date
    	    localtime_r(&blocks[i].planned_from.tv_sec,&temp_tm);strftime(buf1,sizeof(buf1),STRF_DATETIME,&temp_tm);
	    localtime_r(&blocks[i].planned_to.tv_sec,&temp_tm);strftime(buf2,sizeof(buf2),STRF_DATETIME,&temp_tm);
		
	    // print block header
	    printf("%4d: \"%s\"->\"%s\" [%4.2lf][%3d] '%s'\n",i,buf1,buf2,blocks[i].duration,blocks[i].events_count,blocks[i].title);
		
	    for(int j=0;j<blocks[i].events_count;j++)
	        printf
	        (
	    	    "\t%3d\t%s,%8d,%8d,\"%s\"\n",
		    j,
		    blocks[i].events[j].id,
		    blocks[i].events[j].tc,
		    blocks[i].events[j].dur,
		    blocks[i].events[j].title
		);

	};
    };
};

/*
		// reset counter
		nsec_counter = 0;


*/
static int schd_file_load(struct instance_item* instance)
{
    long i,j,d;

    // try to lock mutex for update
    if(pthread_mutex_trylock(&instance->lock_block_update) != 0)
	// perhaps busy - return
	return 0;
	
    
    // lock readers
    pthread_mutex_lock(&instance->lock_block_read);
		
    // try to load schedule
		
    // reset objs
    struct schd_block_item* blocks = 0;
    long blocks_count = 0;


    // try to open file
    FILE* f = fopen
    (
	(instance->cfg_schd_blocks_aux)		// if aux mode is used
	?
	instance->cfg_schd_blocks_file_aux	// use aux file
	:
	instance->cfg_schd_blocks_file, 	// else use default file
	"r+t"
    );

    // check if unable to open
    if (!(f))
    {
	// unable to open file
	pthread_mutex_unlock(&instance->lock_block_read);
	pthread_mutex_unlock(&instance->lock_block_update);
	syslog(LOG_NOTICE, "schd_file_load: Error in opening file '%s'", instance->cfg_schd_blocks_file );	
	return 0;
    };
    
    
    // load blocks
    blocks_count = schd_build(blocks,f);

    // close file
    fclose(f);

    // notify
    syslog(LOG_NOTICE, "loop_schd: blocks file reloaded [%d blocks]", blocks_count );
		    
//DEBUG!
//schd_print(blocks,blocks_count);

	    
    // swap instance block data to loaded
		    
    // free old
    if(instance->blocks)
	schd_free(instance->blocks,instance->blocks_count);
			
    // assign new
    instance->blocks = blocks;
    instance->blocks_count = blocks_count;
    instance->current_block = NULL;
		    
    // find new current block
		    
    // get current datetime
    struct timeval now;
    gettimeofday(&now, NULL);

    // current block		    
    for(i=0,j=(-1),d=0;(i<blocks_count)&&(j == (-1));i++)
	if
	(
	    ( blocks[i].planned_from.tv_sec  <= now.tv_sec )
	    &&
	    ( now.tv_sec <= blocks[i].planned_to.tv_sec  )
	    &&
	    (!(blacklist_blocks_present(blocks[i].sys_id)))
	)
	    j=i;


    if(j == (-1))
	// continue search for nearest start time
	for(i=0,j=(-1),d=0;(i<blocks_count)&&(j == (-1));i++)
	    if( (blocks[i].planned_from.tv_sec + 5) >= now.tv_sec )
	    {
		// block in nearest future
			    
		// how is near
		if
		(
		    (j == (-1))
		    &&
		    (!(blacklist_blocks_present(blocks[i].sys_id)))
		)
		{
		    j = i;
		    d = blocks[i].planned_from.tv_sec + 5 - now.tv_sec;
		}
		else
		{
		    if
		    (
			(d > (blocks[i].planned_from.tv_sec + 5 - now.tv_sec))
			&&
			(!(blacklist_blocks_present(blocks[i].sys_id)))
		    )
		    {
		        j = i;
		        d = blocks[i].planned_from.tv_sec + 5 - now.tv_sec;
		    };
		};
	    };
		    
    // check what we found
    if(j == (-1))
	syslog(LOG_NOTICE, "loop_schd: no current blocks found");
    else
    {
	syslog(LOG_NOTICE, "loop_schd: current block is %d (sys_id='%d', title='%s')", j, blocks[j].sys_id, blocks[j].title );
	instance->current_block = &blocks[j];
    };
		    

    // unlock all mutexes
    pthread_mutex_unlock(&instance->lock_block_read);
    pthread_mutex_unlock(&instance->lock_block_update);

    return 1;
};


static int schd_file_is_modified(struct instance_item* instance)
{
    char* file2analize = 
	(instance->cfg_schd_blocks_aux)		// if aux mode is used
	?
	instance->cfg_schd_blocks_file_aux	// use aux file
	:
	instance->cfg_schd_blocks_file; 	// else use default file
    

    if(stat(file2analize, &instance->schd_file_stat_current) == 0)
    {
	if
	(
	    (instance->schd_file_stat_prev.st_mtime != instance->schd_file_stat_current.st_mtime)
	    ||
	    (instance->schd_file_stat_prev.st_ctime != instance->schd_file_stat_current.st_ctime)
	)
	    return 1;
    };
    
    return 0;
};

static void* loop_schd(void* p)
{
    struct timespec sleep_time; sleep_time.tv_sec = 0; sleep_time.tv_nsec =  300*1000;
    struct timespec slept_time;

    // cast instance ctrl struct
    struct instance_item* instance = (struct instance_item*)p;
    long nsec_counter = (1000 * instance->cfg_schd_reload_interval);
    
    // setup flag thread active
    instance->f_schd = 1;
    syslog(LOG_NOTICE, "loop_schd: started");
    
    // loop started
    while((instance->f_schd)&&(!(instance->f_exit)))
    {
	// detect if file modified
	if((schd_file_is_modified(instance)) && (instance->state != SCHD_POSTROLL))
	{
	    if(schd_file_load(instance))
	    {
		// sync stat structs
		instance->schd_file_stat_prev = instance->schd_file_stat_current;
	    };
	};

	nanosleep(&sleep_time,&slept_time);
	
    };
    
    // return value
    syslog(LOG_NOTICE, "loop_schd: finished");
    pthread_exit(0);

    return NULL;    
};


int stop_schd(struct instance_item* instance)
{
    void* loop_schd_exit;

    if(instance->f_schd)
    {
	syslog(LOG_NOTICE, "stop_schd: Waiting for 'playlist reloader' thread exit");

        instance->f_schd = 0;

        if(pthread_join(instance->th_schd, &loop_schd_exit) < 0)
	    syslog(LOG_NOTICE, "stop_schd: Something wrong with joining to 'playlist reloader' thread");
	else
	    syslog(LOG_NOTICE, "stop_schd: 'playlist reloader' exits with '0x%.8X' ", (long)loop_schd_exit);
    }
    else
        syslog(LOG_NOTICE, "stop_schd: 'playlist reloader' seems allready finished or not runned");

    free(blacklist_blocks_sys_id);
    blacklist_blocks_count = 0;

    // no thread started - exit
    return 1;
};


int start_schd(struct instance_item* instance)
{
    blacklist_blocks_count = 0;
    blacklist_blocks_sys_id = (long*)malloc(0);


    if(pthread_create(&instance->th_schd, &instance->th_schd_attr, loop_schd, instance) < 0)
    {
	syslog(LOG_ERR, "start_schd: Unable to start 'playlist reloader' thread!");
        return 0;
    }
    else
        return 1;
};




void blacklist_blocks_add(long sys_id)
{
    // realloc
    blacklist_blocks_sys_id = (long*)realloc(blacklist_blocks_sys_id, sizeof(long) * (blacklist_blocks_count + 1));
    blacklist_blocks_sys_id[blacklist_blocks_count] = sys_id;
    blacklist_blocks_count++;
    syslog(LOG_ERR, "blacklist_blocks_add: added #%d", sys_id);
    
};

long blacklist_blocks_present(long sys_id)
{
    for(int i=0;i<blacklist_blocks_count;i++)
	if(blacklist_blocks_sys_id[i] == sys_id)
	    return 1;
    return 0;
};

/*

int main(void)
{

    return 0;
};

*/
