/*

Changelog:
    2005-11-17:
	*Delaying stop/start of extcmd params added

    2005-11-11:
	*Threads priority adjastment. Setting initial attributes to threads.

    2005-11-09:
	*remember played block functionality added.
	
    2005-11-07:
	*LOG_LOCAL0 - LOG_LOCAL7 log output fixed.

    2005-11-04 04:00:00
	*Seems model is working !!! Let's test and fix!
	
*/


#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <syslog.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <ncurses.h>

#include "schd_replacer.h"
#include "dvout.h"
#include "schd.h"
#include "display.h"
#include "statectl.h"
#include "cacheman.h"
#include "switcher.h"
#include "extcmd.h"

static struct instance_item prg;


static void init_instance_threads(struct instance_item* instance)
{
    // init attr structs
    pthread_attr_init(&instance->th_dvout_attr);
    pthread_attr_init(&instance->th_schd_attr);
    pthread_attr_init(&instance->th_display_attr);
    pthread_attr_init(&instance->th_statectl_attr);
    pthread_attr_init(&instance->th_cacheman_attr);
    pthread_attr_init(&instance->th_switcher_attr);
    pthread_attr_init(&instance->th_extcmd_attr);
    
    // setup scheduling polici
    pthread_attr_setschedpolicy(&instance->th_dvout_attr,	SCHED_RR);
    pthread_attr_setschedpolicy(&instance->th_schd_attr,	SCHED_FIFO);
    pthread_attr_setschedpolicy(&instance->th_display_attr,	SCHED_OTHER);
    pthread_attr_setschedpolicy(&instance->th_statectl_attr,	SCHED_RR);
    pthread_attr_setschedpolicy(&instance->th_cacheman_attr,	SCHED_OTHER);
    pthread_attr_setschedpolicy(&instance->th_switcher_attr,	SCHED_FIFO);
    pthread_attr_setschedpolicy(&instance->th_extcmd_attr,	SCHED_FIFO);
};


static long log_nos[] = {LOG_LOCAL0,LOG_LOCAL1,LOG_LOCAL2,LOG_LOCAL3,LOG_LOCAL4,LOG_LOCAL5,LOG_LOCAL6,LOG_LOCAL7};

// handle ctrl-c
static sighandler_t sig_int;
static void int_handler(int signal)
{
    syslog(LOG_NOTICE,"Interrupt signal recieved!");
    prg.f_exit = 1;
};

// options
static const char short_options [] = "d:r:h";
static const struct option long_options [] =
{
    { "dv-dev-name",		required_argument,      NULL,   'd'},
    { "dv-ringbuffers",		required_argument,      NULL,   'r'},
    { "reload-interval",	required_argument,      NULL,   'i'},
    { "blocks-file",		required_argument,      NULL,   'b'},
    { "blocks-file-aux",	required_argument,      NULL,   'a'},
    { "display-refresh",	required_argument,      NULL,   'u'},

    { "switch-in-delay",	required_argument,      NULL,   '1'},
    { "switch-out-delay",	required_argument,      NULL,   '2'},
    { "cache-buffer-length",	required_argument,      NULL,   'c'},
    { "video-path",		required_argument,      NULL,   'p'},
    { "switcher-dev-name",	required_argument,      NULL,   's'},
    { "extcmd-dev-name",	required_argument,      NULL,   'e'},
    { "instance-no",		required_argument,      NULL,   '3'},
    { "log-no",			required_argument,      NULL,   '4'},
    { "remember-played-block",	required_argument,      NULL,   'm'},
    { "extcmd-delay-stop",	required_argument,      NULL,   '5'},
    { "extcmd-delay-play",	required_argument,      NULL,   '6'},


    { "help",           	no_argument,            NULL,   'h'},
    { 0,                	0,                      0,      0}
};

// setup default params 
#define DEF_DV_DEV_NAME "/dev/dv1394_PAL_out_0"
#define DEF_DV_RINGBUFFERS 3
#define DEF_SCHD_RELOAD_INTERVAL 10
#define DEF_SCHD_BLOCKS_FILE "/home/verem/repl_sched/schd_replacer.schedule"
#define DEF_SCHD_BLOCKS_FILE_AUX "/home/verem/repl_sched/schd_replacer.schedule.aux"
#define DEF_DISPLAY_REFRESH_INTERVAL 1
#define DEF_STATECTL_IN_DELAY 100
#define DEF_STATECTL_OUT_DELAY 100
#define DEF_CACHEMAN_BUFFER_LENGTH 250
#define DEF_CACHEMAN_VIDEO_PATH "/home/verem/video"
#define DEF_SWITCHER_DEV_NAME "/dev/ttyS0"
#define DEF_EXTCMD_DEV_NAME "/dev/ttyS1"
#define DEF_INSTANCE_NO 0
#define DEF_LOG_NO 6
#define DEF_REMEMBER_PLAYED_BLOCK 0
#define DEF_EXTCMD_DELAY_STOP 0
#define DEF_EXTCMD_DELAY_PLAY 0


static void set_defaults(struct instance_item* instance)
{
    // clear struct
    memset(instance,0,sizeof(struct instance_item));

    // init mutexes
    pthread_mutex_init(&instance->lock_block_update, NULL);
    pthread_mutex_init(&instance->lock_block_read, NULL);
    
    // setup params    
    strcpy(instance->cfg_dv_dev_name,DEF_DV_DEV_NAME);
    instance->cfg_dv_ringbuffers = DEF_DV_RINGBUFFERS;
    instance->cfg_schd_reload_interval = DEF_SCHD_RELOAD_INTERVAL;
    strcpy(instance->cfg_schd_blocks_file, DEF_SCHD_BLOCKS_FILE);
    strcpy(instance->cfg_schd_blocks_file_aux, DEF_SCHD_BLOCKS_FILE_AUX);
    instance->cfg_display_refresh_interval = DEF_DISPLAY_REFRESH_INTERVAL;
    instance->cfg_statectl_out_delay = DEF_STATECTL_OUT_DELAY;
    instance->cfg_statectl_in_delay = DEF_STATECTL_IN_DELAY;
    instance->cfg_cacheman_buffer_length = DEF_CACHEMAN_BUFFER_LENGTH;
    strcpy(instance->cfg_cacheman_video_path,DEF_CACHEMAN_VIDEO_PATH);
    strcpy(instance->cfg_switcher_device_name, 	DEF_SWITCHER_DEV_NAME);
    strcpy(instance->cfg_extcmd_device_name, 	DEF_EXTCMD_DEV_NAME);
    instance->cfg_instance_no = DEF_INSTANCE_NO;
    instance->cfg_log_no = DEF_LOG_NO;
    instance->cfg_schd_remember_played_block = DEF_REMEMBER_PLAYED_BLOCK;

    instance->cfg_extcmd_delay_stop = DEF_EXTCMD_DELAY_STOP;
    instance->cfg_extcmd_delay_stop = DEF_EXTCMD_DELAY_PLAY;

    
    // ext devs and cmds enables
    instance->f_enable_switcher = 1;
    instance->f_enable_extcmd = 1;
    
    // init threads
    init_instance_threads(instance);

};

// usage information
static void usage(void)
{
    printf
    (
        "Usage: psvbid <args>\n"
        "args:\n"
        "\t--dv-dev-name=<devname>\tdv1394 device name (mknod -m 666 /dev/dv1394/0 c 171 35) (default: '%s')\n"
        "\t--dv-ringbuffers=N\tnumber of buffers allocated for White level of signal (default %ld)\n"
        "\t--reload-interval=N\tnumber of seconds before schedule reload (default %ld)\n"
        "\t--blocks-file=<filename>\tfile containes schedule of blocks (default '%s')\n"
        "\t--blocks-file-aux=<filename>\tfile containes auxilarity schedule of blocks (default '%s')\n"
        "\t--switch-in-delay=N\tdelay in N miliseconds 'in' switcher operations (default '%ld')\n"
        "\t--switch-out-delay=N\tdelay in N miliseconds 'out' switcher operations (default '%ld')\n"
        "\t--cache-buffer-length=N\tnumber of frames to cache (default '%ld')\n"
        "\t--video-path=<path>\tpath to directory where video stored (default '%d')\n"
        "\t--switcher-dev-name=<path>\texternal switcher/keyer serial device name (default '%s')\n"
        "\t--extcmd-dev-name=<path>\texternal command detector serial device name (default '%s')\n"
        "\t--instance-no=N\tnumber or running application (default sch_replacer'%d')\n"
        "\t--log-no=N\tsyslog param LOG_LOCAL'N' (default LOG_LOCAL'%d')\n"
        "\t--remember-played-block=0|1\tremember played block, prevents cueing played block (default '%d')\n"

        "\t--extcmd-delay-stop=N\tDelay in miliseconds of 'stop' command in extcmd module (default '%d')\n"
        "\t--extcmd-delay-play=N\tDelay in miliseconds of 'play' command in extcmd module (default '%d')\n"

        "\t--help\tShow program usage\n",
        DEF_DV_DEV_NAME,
        DEF_DV_RINGBUFFERS,
	DEF_SCHD_RELOAD_INTERVAL,
	DEF_SCHD_BLOCKS_FILE,
	DEF_SCHD_BLOCKS_FILE_AUX,
	DEF_STATECTL_IN_DELAY,
	DEF_STATECTL_OUT_DELAY,
	DEF_CACHEMAN_BUFFER_LENGTH,
	DEF_CACHEMAN_VIDEO_PATH,
	DEF_SWITCHER_DEV_NAME,
	DEF_EXTCMD_DEV_NAME,
	DEF_INSTANCE_NO,
	DEF_LOG_NO,
	DEF_REMEMBER_PLAYED_BLOCK,
	DEF_EXTCMD_DELAY_STOP,
	DEF_EXTCMD_DELAY_PLAY
    );
};


int main(int argc, char** argv)
{
    int i,c,index;
    
    // setup defaults
    set_defaults(&prg);
    
    // program arguments processing
    while(1)
    {
        c = getopt_long (argc, argv, short_options, long_options, &index);

        if(c == (-1)) break;

        switch(c)
        {
            case 0: break;
            case 'd': strncpy(prg.cfg_dv_dev_name,optarg,sizeof(prg.cfg_dv_dev_name) - 1);break;
            case 'r': prg.cfg_dv_ringbuffers = atol(optarg);break;
            case 'u': prg.cfg_display_refresh_interval = atol(optarg);break;
            case 'i': prg.cfg_schd_reload_interval = atol(optarg);break;
            case 'b': strncpy(prg.cfg_schd_blocks_file,optarg,sizeof(prg.cfg_schd_blocks_file) - 1);break;
            case 'a': strncpy(prg.cfg_schd_blocks_file_aux,optarg,sizeof(prg.cfg_schd_blocks_file_aux) - 1);break;
            case '1': prg.cfg_statectl_in_delay = atol(optarg);break;
            case '2': prg.cfg_statectl_out_delay = atol(optarg);break;
            case 'c': prg.cfg_cacheman_buffer_length = atol(optarg);break;
            case 'p': strncpy(prg.cfg_cacheman_video_path,optarg,sizeof(prg.cfg_cacheman_video_path) - 1);break;
            case 's': strncpy(prg.cfg_switcher_device_name,optarg,sizeof(prg.cfg_switcher_device_name) - 1);break;
            case 'e': strncpy(prg.cfg_extcmd_device_name,optarg,sizeof(prg.cfg_extcmd_device_name) - 1);break;
            case '3': prg.cfg_instance_no = atol(optarg);break;
            case '4': prg.cfg_log_no = atol(optarg);break;
            case 'm': prg.cfg_schd_remember_played_block = atol(optarg);break;
	    case '5': prg.cfg_extcmd_delay_stop = atol(optarg);break;
	    case '6': prg.cfg_extcmd_delay_play = atol(optarg);break;
	
            case 'h': usage();exit(EXIT_FAILURE);break;
            default:
                printf("ERROR: Incorrect argument!\n");
                usage();
                exit (EXIT_FAILURE);
        };
    };

    // open syslog 
    char ident[128]; sprintf(ident,"schd_replacer%d",prg.cfg_instance_no);
    openlog(ident,LOG_PID,  ( (prg.cfg_log_no < 8)?log_nos[prg.cfg_log_no]:LOG_LOCAL0) );
    syslog(LOG_NOTICE,"schd_replacer.h " VERSION " started");

    // set new int handler
    sig_int = signal(SIGINT, int_handler);


    // start services
    if
    (
	(start_cacheman(&prg))
	&&
	(start_dvout(&prg))
	&&
	(start_schd(&prg))
	&&
	(start_display(&prg))
	&&
	(start_statectl(&prg))
	&&
	(start_switcher(&prg))
	&&
	(start_extcmd(&prg))
    )
        // wait for prg interrupt
	while(!(prg.f_exit)) sleep(1);
    
    // finish all threads and services
    stop_statectl(&prg);
    stop_schd(&prg);
    stop_dvout(&prg);
    stop_display(&prg);
    stop_cacheman(&prg);
    stop_switcher(&prg);
    stop_extcmd(&prg);
    
    /* report about transaction */
    syslog(LOG_NOTICE,"program finished");

    /* close log */
    closelog();

    // exit
    exit(EXIT_SUCCESS);

    return 0;
};


