#ifndef schd_replacer_h
#define schd_replacer_h

#define VERSION "1.0-rc5-20051117"

#include <pthread.h>
#include <sys/stat.h>

#include "schd.h"

#define MAX_FILE_LEN 1024

enum SCHD_STATES
{
    SCHD_IDLE = 0,
    SCHD_CUE_INIT,
    SCHD_CUE_DONE,
    SCHD_PLAYBACK,
    SCHD_POSTROLL
} ;

struct instance_item
{
    // sys state
    SCHD_STATES state;
    struct timeval state_change;
    
    // threads ctrls
    pthread_t th_dvout;
    pthread_attr_t th_dvout_attr;
    long f_dvout;
    pthread_t th_schd;
    pthread_attr_t th_schd_attr;
    long f_schd;
    pthread_t th_display;
    pthread_attr_t th_display_attr;
    long f_display;
    pthread_t th_statectl;
    pthread_attr_t th_statectl_attr;
    long f_statectl;
    pthread_t th_cacheman;
    pthread_attr_t th_cacheman_attr;
    long f_cacheman;
    pthread_t th_switcher;
    pthread_attr_t th_switcher_attr;
    long f_switcher;
    pthread_t th_extcmd;
    pthread_attr_t th_extcmd_attr;
    long f_extcmd;

    // flags
    long f_exit;
    long f_cue_init;
    long f_cue_done;
    long f_switch_TO_ME;
    long f_switch_OUT_ME;
    long f_stop;
    long f_stop_emerg;
    long f_play;
    long f_force_cue;
    long f_loop_block;
    long f_enable_switcher;
    long f_enable_extcmd;

    // vars
    struct timeval f_stop_occur;
    struct timeval f_stop_emerg_occur;
    struct timeval f_play_occur;
    struct timeval f_switch_cmd;
    struct timeval f_switch_ack;
    struct timeval f_extcmd_cmd;
    struct timeval f_extcmd_ack;
    long playback_remain;
    long playback_count;
    long* playback_remains;
    long* playback_lengths;
    long playback_length;
    long cacheman_frames_loaded;
    long cacheman_frames_sent;
    long dv_status;
    char *dv_state;
    struct stat schd_file_stat_prev;
    struct stat schd_file_stat_current;


    // plan
    struct schd_block_item* blocks;
    struct schd_block_item* current_block;
    long blocks_count;
    pthread_mutex_t lock_block_update;
    pthread_mutex_t lock_block_read;

    // CONFIGS
    char cfg_dv_dev_name[MAX_FILE_LEN];
    long cfg_dv_ringbuffers;
    long cfg_schd_reload_interval;
    long cfg_schd_remember_played_block;
    char cfg_schd_blocks_file[MAX_FILE_LEN];
    char cfg_schd_blocks_file_aux[MAX_FILE_LEN];
    long cfg_schd_blocks_aux;
    long cfg_display_refresh_interval;    
    long cfg_statectl_out_delay;
    long cfg_statectl_in_delay;
    long cfg_cacheman_buffer_length;
    char cfg_cacheman_video_path[MAX_FILE_LEN];
    char cfg_switcher_device_name[MAX_FILE_LEN];
    long cfg_instance_no;
    long cfg_log_no;
    char cfg_extcmd_device_name[MAX_FILE_LEN];

    long cfg_extcmd_delay_stop;
    long cfg_extcmd_delay_play;
    
};


#endif
