/*

Changelog:
    2005-11-11:
	*Threads priority adjastment.

    2005-11-09:
	*'cfg_schd_remember_played_block'
	*'cfg_schd_blocks_aux'
	
    2005-11-05:
	*Force CUE operations added.
	*Smart schedule file load fixups - resetting schd_file_stat_prev 
	during POSTROLL

    2005-11-04 04:00:00
	*Seems model is working !!! Let's test and fix!
	
*/


#include <unistd.h>
#include <stdio.h>
#include <syslog.h>
#include <time.h>
#include <string.h>

#include "schd_replacer.h"
#include "statectl.h"

// state ctrl
static void* loop_statectl(void* p)
{
    long i;
    
    // cast instance ctrl struct
    struct instance_item* instance = (struct instance_item*)p;

    // setup flag thread active
    instance->f_statectl = 1;
    syslog(LOG_NOTICE, "loop_statectl: started");

    struct timespec sleep_time; sleep_time.tv_sec = 0; sleep_time.tv_nsec =  1000;
    struct timespec slept_time;
    
    // keep prev state
    SCHD_STATES prev_state = instance->state;
    SCHD_STATES new_state = instance->state;
    struct timeval current_time;
    
    while((instance->f_statectl)&&(!(instance->f_exit)))
    {
	// get current time
	gettimeofday(&current_time,NULL);
	
	// default new state
	new_state = instance->state;
	
	// depends on state
	pthread_mutex_lock(&instance->lock_block_read);
	
	// check for forbidden moves
	//
	//
	//else
	switch(instance->state)
	{
	    case SCHD_IDLE:	// system in idle state
		// reset force cue if previous state was not IDLE
		if(prev_state != new_state)
		    instance->f_force_cue = 0;
		else
		    if (instance->f_force_cue)
		    {
			// first reset this flag
			instance->f_force_cue = 0;
			
			// fix FROM time of current block
			if(instance->current_block)
			    instance->current_block->planned_from.tv_sec = current_time.tv_sec - 5;
		    };
		
	    
		if(instance->current_block)
		{
		    // check if block should be cued
		    if
		    (
			(instance->current_block->planned_from.tv_sec < current_time.tv_sec)
			&&
			(instance->current_block->planned_to.tv_sec > current_time.tv_sec)
		    )
		    {
			// setup flag
			instance->f_cue_init = 1;
			instance->f_cue_done = 0;
			new_state = SCHD_CUE_INIT;
			syslog(LOG_NOTICE, "loop_statectl: SCHD_IDLE->SCHD_CUE_INIT at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
		    };
		}
		break;

	    case SCHD_CUE_INIT:	// system cuing videos
		if(instance->f_cue_done)
		{
		    // cue done - lets go next stage
		    new_state = SCHD_CUE_DONE;
		    instance->f_play = 0;
		    instance->f_stop = 0;
		    instance->f_stop_emerg = 0;
		    instance->f_switch_TO_ME = 0;
		    instance->f_switch_OUT_ME = 0;
		    syslog(LOG_NOTICE, "loop_statectl: SCHD_CUE_INIT->SCHD_CUE_DONE at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
		}
		else
		{
		    // check if flag was reseted
		    if(!(instance->f_cue_init))
		    {
			new_state = SCHD_POSTROLL;
			syslog(LOG_NOTICE, "loop_statectl: SCHD_CUE_INIT->SCHD_POSTROLL at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
		    };
		};
		break;

	    case SCHD_CUE_DONE:	// system cued videos and ready for playout
		// check if time is expired? or emergency stop occurs
		if
		(
		    (instance->current_block->planned_to.tv_sec < current_time.tv_sec)
		    ||
		    (instance->f_stop_emerg)
		)
		{
		    // xpires
		    new_state = SCHD_POSTROLL;
		    syslog(LOG_NOTICE, "loop_statectl: SCHD_CUE_DONE->POSTROLL at %ld.%.6ld due to expire",current_time.tv_sec,current_time.tv_usec);
		}
		else
		// check if command to PLAY
		if(instance->f_play)
		{
		    // command play accepted !!!
		    new_state = SCHD_PLAYBACK;
		    instance->f_stop = 0;
		    instance->f_stop_emerg = 0;
		    instance->f_play = 0;
		    instance->f_switch_TO_ME = 0;
		    instance->f_switch_OUT_ME = 0;
		    syslog(LOG_NOTICE, "loop_statectl: SCHD_CUE_DONE->SCHD_PLAYBACK at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
		}
		break;

	    case SCHD_PLAYBACK:	// system in playback state - we can stop switch in switch out
		// emergency stop
		if(instance->f_stop_emerg)
		{
		    new_state = SCHD_POSTROLL;
		    instance->f_switch_OUT_ME = 1;
		    syslog(LOG_NOTICE, "loop_statectl: SCHD_PLAYBACK->POSTROLL at %ld.%.6ld [EMERGENCY]",current_time.tv_sec,current_time.tv_usec);
		    break;
		}
	    
		// check for rized stop flag - delayed out
		if
		(
		    (instance->f_stop)
		    &&
		    (timeval_diff_mili(current_time,instance->f_stop_occur) > instance->cfg_statectl_out_delay )
		)
		{
		    new_state = SCHD_POSTROLL;
		    instance->f_switch_OUT_ME = 1;
		    syslog(LOG_NOTICE, "loop_statectl: SCHD_PLAYBACK->POSTROLL at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
		    break;
		};
	    
		// check for rizing TO ME FLAG
		if
		(
		    (!(instance->f_switch_TO_ME))
		    &&
		    (timeval_diff_mili(current_time,instance->state_change) > instance->cfg_statectl_in_delay )
		)
		{
		    instance->f_switch_TO_ME = 1;
		    syslog(LOG_NOTICE, "loop_statectl: 'cfg_statectl_in_delay' expires - rize 'f_switch_TO_ME' at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
		};		
		
		// check reel finished
		if
		(
		    (instance->playback_remain == 0)
		    &&
		    (!(instance->f_stop))
		)
		{
		    instance->f_stop = 1;
		    gettimeofday(&instance->f_stop_occur,NULL);
		    syslog(LOG_NOTICE, "loop_statectl: SCHD_PLAYBACK out of reel detected - rize 'f_stop' at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
		};
		
		break;
		
	    case SCHD_POSTROLL:
		// reset schedule ptr - force reload
		memset(&instance->schd_file_stat_prev,0,sizeof(instance->schd_file_stat_prev));
		
		// just check if delay elapsed !
		if(timeval_diff_mili(current_time,instance->state_change) > 5000)
		{
		    new_state = SCHD_IDLE;
		    syslog(LOG_NOTICE, "loop_statectl: POSTROLL->IDLE at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
		    
		    // reset current playlist !!!!
		    if(instance->current_block)
		    {
			// forget about current block
			if
			(
			    (!(instance->f_stop_emerg))				// not emerg exit
			    &&							// and
			    (instance->cfg_schd_remember_played_block)		// option active
			)
			{
			    if(!(instance->cfg_schd_blocks_aux))		// do not mark blocks from AUX list
				blacklist_blocks_add(instance->current_block->sys_id);
			};
		    
			instance->current_block = NULL;
		    };
		};
		break;
				    
	};
	pthread_mutex_unlock(&instance->lock_block_read);
	
	// switch states
	prev_state = instance->state;
	if( instance->state != new_state)
	    gettimeofday(&instance->state_change, NULL);
	instance->state = new_state;
	
	// sleep
 	nanosleep(&sleep_time,&slept_time);
    };
    
    // setup flag thread active
    instance->f_statectl = 0;

    // return value
    syslog(LOG_NOTICE, "loop_statectl: finished");
    pthread_exit(0);
    
    return NULL;
};

int stop_statectl(struct instance_item* instance)
{
    void* loop_statectl_exit;
    
    if(instance->f_statectl)
    {
    	syslog(LOG_NOTICE, "stop_statectl: Waiting for 'state controller' thread exit");
    	instance->f_statectl = 0;
    	if(pthread_join(instance->th_statectl, &loop_statectl_exit) < 0)
	    syslog(LOG_NOTICE, "stop_statectl: Something wrong with joining to 'dv_out' thread");
	else
	    syslog(LOG_NOTICE, "stop_statectl: 'state controller' exits with '0x%.8X' ", (long)loop_statectl_exit);
    }
    else
	syslog(LOG_NOTICE, "stop_statectl: 'state controller' seems allready finished or not runned");
    	// no thread started - exit
        return 1;
};

int start_statectl(struct instance_item* instance)
{
    if(pthread_create(&instance->th_statectl, &instance->th_statectl_attr, loop_statectl, instance) < 0)
    {
	syslog(LOG_ERR, "start_statectl: Unable to start 'dv_out' thread!");
	return 0;
    }
    else
	return 1;
};

long timeval_diff_mili(struct timeval &A, struct timeval &B)
{
    return ( (A.tv_sec - B.tv_sec)*1000000 + A.tv_usec - B.tv_usec)/1000;
};

