/*

Changelog:
    2005-11-11:
	*Threads priority adjustment.

    2005-11-06:
	*Enabling/Disabling control of switcher is added. Command to switcher sent only
	in case that 'f_enable_switcher' flag is rized.

    2005-11-04 04:00:00
	*Seems model is working !!! Let's test and fix!
	
*/


#include <unistd.h>
#include <stdio.h>
#include <syslog.h>
#include <time.h>
#include <fcntl.h>
#include <termios.h>
#include <strings.h>

#include "schd_replacer.h"

static struct termios oldtio,newtio;
static int serial_handler;


/*
    M2100 switcher control protocol 


IN1(M1 main) -> PROGRAM - "OUT ME"
C:\temp\airforce>TSTmym2100.exe COM2 CMD XPT_TAKE 1 2 0
[ 12]>02 00 08 01 00 05 06 00 01 02 00 E9
[  0]<

IN2(DV stream) -> PROGRAM - "TO ME"
C:\temp\airforce>TSTmym2100.exe COM2 CMD XPT_TAKE 1 3 0
[ 12]>02 00 08 01 00 05 06 00 01 03 00 E8
[  0]<

*/

static unsigned char switcher_cmd_datas[2][20] = 
{
    {0x02, 0x00, 0x08, 0x01, 0x00, 0x05, 0x06, 0x00, 0x01, 0x03, 0x00, 0xE8},	// TO ME
    {0x02, 0x00, 0x08, 0x01, 0x00, 0x05, 0x06, 0x00, 0x01, 0x02, 0x00, 0xE9}	// OUT ME
};

static long switcher_cmd_lengths[] = {12,12};

// switcher loop
static void* loop_switcher(void* p)
{
    // cast instance ctrl struct
    struct instance_item* instance = (struct instance_item*)p;

    // setup flag thread active
    instance->f_switcher = 1;
    syslog(LOG_NOTICE, "loop_switcher: started");

    //
    struct timeval current_time;

    struct timespec sleep_time;
    sleep_time.tv_sec = 0; sleep_time.tv_nsec =  100;
    struct timespec slept_time;

    // trigger ctrl
    long f_switch_TO_ME = 0;
    long f_switch_OUT_ME = 0;
    void *switcher_cmd_data = NULL;
    long switcher_cmd_length = 0;

    // cmd sending logging
    gettimeofday(&instance->f_switch_cmd,NULL);
    gettimeofday(&instance->f_switch_ack,NULL);

    
    while((instance->f_switcher)&&(!(instance->f_exit)))
    {
        // get current time
	gettimeofday(&current_time,NULL);
                
    	// detect rize of f_switch_TO_ME
    	if
    	(
	    (f_switch_TO_ME == 0)
	    &&
	    (instance->f_switch_TO_ME == 1)
	)
	{
	    switcher_cmd_data = switcher_cmd_datas[0];
	    switcher_cmd_length = switcher_cmd_lengths[0];
	    syslog(LOG_NOTICE, "loop_switcher: rise of 'f_switch_TO_ME' detected at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
	};

    	// detect rize of f_switch_OUT_ME
    	if
    	(
	    (f_switch_OUT_ME == 0)
	    &&
	    (instance->f_switch_OUT_ME == 1)
	)
	{
	    switcher_cmd_data = switcher_cmd_datas[1];
	    switcher_cmd_length = switcher_cmd_lengths[1];
	    syslog(LOG_NOTICE, "loop_switcher: rise of 'f_switch_OUT_ME' detected at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
	};

	// should we send datas
	if((switcher_cmd_data)&&(instance->f_enable_switcher))
	{
	    // check if device open
	    if (serial_handler > 0)
	    {
	    	// discard all data in buffers
	    	tcflush(serial_handler, TCIOFLUSH);
	    	
	    	// send data
	    	write(serial_handler, switcher_cmd_data, switcher_cmd_length);
	    	gettimeofday(&current_time,NULL);
		gettimeofday(&instance->f_switch_cmd,NULL);
	    	syslog(LOG_NOTICE, "loop_switcher: data sent to serial device at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);

                // read data from device
		fd_set fds;
		struct timeval tv;
                int r;
                int reply = 0;
                // async read
                FD_ZERO (&fds);
                FD_SET (serial_handler, &fds);
                tv.tv_sec = 0; /* Timeout 'sec'. */
                tv.tv_usec = 80*1000; /* Timeout 'usec'. TWO FRAMES WAIT (hardcoded)  */
                r = select (serial_handler + 1, &fds, NULL, NULL, &tv);
	    	gettimeofday(&current_time,NULL);
                if(r>0)
                {
                    // read data
		    gettimeofday(&instance->f_switch_ack,NULL);
		    read(serial_handler, &reply, 1);
		    syslog(LOG_NOTICE, "loop_switcher: data recieved from serial device [0x%.8X] at %ld.%.6ld",reply,current_time.tv_sec,current_time.tv_usec);
	        }
	        else
		    syslog(LOG_NOTICE, "loop_switcher: data NOT recieved from serial device due to timeout or error at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
        
	    }
	    else
	        syslog(LOG_NOTICE, "loop_switcher: unable to send cmd due to serial devices close at %ld.%.6ld",current_time.tv_sec,current_time.tv_usec);
	};
	

	// reset datas
	switcher_cmd_data = NULL;
	switcher_cmd_length = 0;
	f_switch_TO_ME = instance->f_switch_TO_ME;
	f_switch_OUT_ME = instance->f_switch_OUT_ME;

	// sleep a bit
    	nanosleep(&sleep_time, &slept_time);
    };
		
    
    // setup flag thread active
    instance->f_switcher = 0;

    // return value
    syslog(LOG_NOTICE, "loop_switcher: finished");
    pthread_exit(0);
    
    return NULL;
};

int stop_switcher(struct instance_item* instance)
{
    void* loop_switcher_exit;
    
    if (serial_handler >0)
    {
        // restore serial device params
        tcflush(serial_handler, TCIFLUSH);
        tcsetattr(serial_handler,TCSANOW,&newtio);
    };
    
    
    if(instance->f_switcher)
    {
    	syslog(LOG_NOTICE, "stop_switcher: Waiting for 'switcher controller' thread exit");
    	instance->f_switcher = 0;
    	if(pthread_join(instance->th_switcher, &loop_switcher_exit) < 0)
	    syslog(LOG_NOTICE, "stop_switcher: Something wrong with joining to 'switcher controller' thread");
	else
	    syslog(LOG_NOTICE, "stop_switcher: 'switcher controller' exits with '0x%.8X' ", (long)loop_switcher_exit);
    }
    else
	syslog(LOG_NOTICE, "stop_switcher: 'switcher controller' seems allready finished or not runned");

    // no thread started - exit
    return 1;
};

int start_switcher(struct instance_item* instance)
{
    // open serial device
    serial_handler = 0;
    
    if( (serial_handler = open(instance->cfg_switcher_device_name, O_RDWR | O_NOCTTY ) ) < 0 )
    {
    	syslog(LOG_ERR, "start_switcher: Unable to open '%s'!", instance->cfg_switcher_device_name);
    }
    else
    {
    	syslog(LOG_NOTICE, "start_switcher: Device '%s' opened successful!",instance->cfg_switcher_device_name);
    	
    	// configure serial device
        tcgetattr(serial_handler,&oldtio);        /* save current serial port settings */
        bzero(&newtio, sizeof(newtio));     /* clear struct for new port settings */
        newtio.c_iflag = IGNPAR;            /* ignore bytes with parity errors */
        newtio.c_oflag = 0;                 /* Raw output */
        newtio.c_lflag = 0;//ICANON;        /* enable canonical input */
        newtio.c_cc[VTIME]    = 0;          /* inter-character timer unused */
        newtio.c_cc[VMIN]     = 1;          /* blocking read until 5 chars received */
        newtio.c_cflag =
	    B19200 |
            PARENB |
            PARODD |
            CS8 |
            CREAD |
            CLOCAL;
        tcflush(serial_handler, TCIFLUSH);
        tcsetattr(serial_handler,TCSANOW,&newtio);
    };
                                                                                            


    if(pthread_create(&instance->th_switcher, &instance->th_switcher_attr, loop_switcher, instance) < 0)
    {
	syslog(LOG_ERR, "start_switcher: Unable to start 'switcher controller' thread!");
	return 0;
    }
    else
	return 1;
};
